//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseApp.cc
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <CommonAPIMessages_m.h>
#include <BaseOverlayMessage_m.h>

#include "BaseApp.h"
#include "BootstrapOracleAccess.h"
#include "GlobalStatisticsAccess.h"
#include "UnderlayConfiguratorAccess.h"


BaseApp::~BaseApp()
{
    //...
}

int BaseApp::numInitStages() const
{
    return MAX_STAGE_APP + 1;
}

void BaseApp::initialize(int stage)
{
    if(stage == MIN_STAGE_APP) {
        // fetch parameters
        debugOutput = par("debugOutput");
        onlyCommonAPIMessages = true;

        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        globalStatistics = GlobalStatisticsAccess().get();

	overlay = check_and_cast<BaseOverlay*>(gate("to_lowerTier")->destinationGate()->ownerModule());

        // determine the terminal's transport address
        thisNode.ip = IPAddressResolver().addressOf(parentModule()->parentModule()).get4();
	WATCH(thisNode);
        //thisNode.key = //solution?

        // statistics
        numSent = 0;
        numReceived = 0;
        bytesSent = 0;
        bytesReceived = 0;

        WATCH(numSent);
        WATCH(numReceived);
        WATCH(bytesSent);
        WATCH(bytesReceived);
    }
    if(stage >= MIN_STAGE_APP && stage <= MAX_STAGE_APP)
        initializeApp(stage);
}

void BaseApp::initializeApp(int stage)
{
    // ...
}

// Process messages passed up from the overlay.
void BaseApp::handleMessage(cMessage* msg)
{
    // Process self-messages.
    if(msg->isSelfMessage()) {
	handleTimerEvent(msg);
	return;
    }
    
    if(msg->arrivedOn("from_lowerTier")) {
	// common API
	CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
	if(commonAPIMsg != NULL)
	    handleCommonAPIMessage(commonAPIMsg);
	    
	else if (onlyCommonAPIMessages == false) {
	    numReceived++;
	    bytesReceived += msg->byteLength();
	    handleAppMessage(msg);
	}
    }
    else if(msg->arrivedOn("from_upperTier"))
	handleUpperMessage(msg);
}

void BaseApp::handleTimerEvent(cMessage* msg)
{
    // process self-messages
}

void BaseApp::callRoute(const OverlayKey& key, cMessage* msg, const NodeHandle& hint)
{
    // create route-message (common API)
    KBRroute* routeMsg = new KBRroute();
    routeMsg->setDestKey(key);
    routeMsg->setHint(hint);
    routeMsg->encapsulate(msg);

    routeMsg->setType(KBR_ROUTE);

    send(routeMsg, "to_lowerTier");

    // debug output
    if (debugOutput)
        EV << "(BaseApp::callRoute()) Node " << thisNode.ip << " sent message "
        << id() << "-" << numSent <<  " to destination key "
        << key.toString(16) << "." << endl;

    // count
    numSent++;
    bytesSent += msg->byteLength();
}

void BaseApp::deliver(OverlayKey& key, cMessage* msg)
{
    // deliver...

    delete msg;
}

void BaseApp::forward(OverlayKey* key, cMessage* msg, NodeHandle* nextHopNode)
{
    // usually do nothing
}

void BaseApp::forwardResponse(const OverlayKey& key, cMessage* msg,
			      const NodeHandle& nextHopNode)
{
    OverlayCtrlInfo* ctrlInfo =
	check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

    //create forwardResponse message (common API)
    KBRforward* forwardMsg = new KBRforward();
    forwardMsg->setDestKey(key);
    forwardMsg->setNextHopNode(nextHopNode);
    forwardMsg->setControlInfo(ctrlInfo);
    forwardMsg->encapsulate(msg);

    forwardMsg->setType(KBR_FORWARD_RESPONSE);

    send(forwardMsg, "to_lowerTier");
}

void BaseApp::update(const NodeHandle& node, bool joined)
{
    // update...
}

void BaseApp::handleCommonAPIMessage(CommonAPIMessage* commonAPIMsg)
{
    cMessage* tempMsg = commonAPIMsg->decapsulate();

    // process interface control information
    OverlayCtrlInfo* overlayCtrlInfo =
	dynamic_cast<OverlayCtrlInfo*>(commonAPIMsg->removeControlInfo());
    if(overlayCtrlInfo != NULL) {
	tempMsg->setControlInfo(overlayCtrlInfo);

	if(debugOutput)
	    EV << "(BaseApp) Node " << thisNode.ip << " received message "
	       << " from node " << overlayCtrlInfo->getSrcNode().ip
	       << "." << endl;
    }

    switch (commonAPIMsg->getType()) {

    case KBR_DELIVER:
	{
	    KBRdeliver* apiMsg = dynamic_cast<KBRdeliver*>(commonAPIMsg);
	    OverlayKey key = apiMsg->getDestKey();
	    NodeHandle nextHopNode = thisNode;
	    //first call forward, then deliver
	    forward(&key, tempMsg, &nextHopNode);

	    if(tempMsg != NULL) {
		//if key or nextHopNode is changed send msg back to overlay
		if(key != apiMsg->getDestKey() ||
		   (!nextHopNode.isUnspecified() &&
		    nextHopNode != thisNode)) {
		    forwardResponse(key, tempMsg, nextHopNode);		
		}
		else {
		    numReceived++;
		    bytesReceived += tempMsg->byteLength();
		    deliver(apiMsg->getDestKey(), tempMsg);
		}			
	    }		
	    break;
	}

    case KBR_FORWARD:
	{
	    KBRforward* apiMsg = dynamic_cast<KBRforward*>(commonAPIMsg);
	    OverlayKey key = apiMsg->getDestKey();
	    NodeHandle nextHopNode = apiMsg->getNextHopNode();

	    forward(&key, tempMsg, &nextHopNode);

	    //if message ist not deleted send it back
	    if(tempMsg != NULL) {
		if(nextHopNode == apiMsg->getNextHopNode())
		    //do we need this?
		    nextHopNode = NodeHandle::UNSPECIFIED_NODE;
		forwardResponse(key, tempMsg, nextHopNode);		
	    }
	    break;
	}

    case KBR_UPDATE:
	{
	    KBRupdate* apiMsg = dynamic_cast<KBRupdate*>(commonAPIMsg);
	    if(apiMsg->getJoined() &&
	       ((apiMsg->getNode().ip == thisNode.ip && 
		 thisNode.port == -1 || apiMsg->getNode().port == thisNode.port) ||
		apiMsg->getNode().key == thisNode.key))
		thisNode = apiMsg->getNode();
	    else
		update(apiMsg->getNode(), apiMsg->getJoined());
		
	    break;
	}

    default:
	{
	    delete tempMsg;
	}
    }
    delete commonAPIMsg;
}

void BaseApp::handleUpperMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::handleAppMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::sendMessageToOverlay(cMessage* msg)
{
    numSent++;
    bytesSent += msg->byteLength();

    send(msg, "to_lowerTier");
}

void BaseApp::finish()
{
    // record scalar data
    recordScalar("BaseApp: Sent Messages", numSent);
    recordScalar("BaseApp: Received Messages", numReceived);
    recordScalar("BaseApp: Sent Bytes", bytesSent);
    recordScalar("BaseApp: Received Bytes", bytesReceived);

    finishApp();
}

void BaseApp::finishApp()
{
    // ...
}
