//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseApp.h
 * @author Bernhard Heep
 */

#ifndef __BASEAPP_H_
#define __BASEAPP_H_

#include <omnetpp.h>

#include <BootstrapOracle.h>
#include <BaseOverlay.h>
#include <GlobalStatistics.h>
#include <UnderlayConfigurator.h>
#include <NodeHandle.h>
#include <OverlayKey.h>
#include <InitStages.h>

class CommonAPIMessage;


/**
 * Base class for applications (Tier 1)
 *
 * Base class for applications (Tier 1) that use overlay functionality.
 * provides common API for structured overlays (KBR)
 *
 * @see KBRTestApp
 * @author Bernhard Heep
 */
class BaseApp : public cSimpleModule
{
private:

    void forwardResponse(const OverlayKey& key, cMessage* msg,
			 const NodeHandle& nextHopNode);

    void handleCommonAPIMessage(CommonAPIMessage* commonAPIMsg);

protected:
    /**
     * method to set InitStage
     */
    int numInitStages() const;

    /**
     * initializes base class-attributes
     *
     * @param stage the init stage
     */
    void initialize(int stage);

    /**
     * initializes derived class-attributes
     *
     * @param stage the init stage
     */
    virtual void initializeApp(int stage);

    /**
     * checks for message type and calls corresponding method
     *
     * checks for message type (from overlay or selfmessage) and calls
     * corresponding method
     * like deliver(), forward(), and timer()
     * @param msg the handled message
     */
    void handleMessage(cMessage* msg);

    /**
     * collects statistical data
     */
    void finish();

    /**
     * collects statistical data of derived app
     */
    virtual void finishApp();

    // common API for structured p2p-overlays
    /**
     * Common API function: calls route-method in overlay
     *
     * encapsulates msg into KBRroute message and sends it to the overlay
     * module
     *
     * @param key destination key
     * @param msg message to route
     * @param hint next hop (usually unused)
     */
    void callRoute(const OverlayKey& key, cMessage* msg,
		   const NodeHandle& hint = NodeHandle::UNSPECIFIED_NODE);

    /**
     * Common API function: handles delivered messages from overlay
     *
     * method to handle decapsulated KBRdeliver messages from overlay module,
     * should be overwritten in derived application
     * @param key destination key
     * @param msg delivered message
     */
    virtual void deliver(OverlayKey& key, cMessage* msg);

    /**
     * Common API function: handles messages from overlay to be forwarded
     *
     * method to handle decapsulated KBRdeliver messages from overlay module,
     * should be overwritten in derived application if needed
     * @param key destination key
     * @param msg message to forward
     * @param nextHopNode next hop
     */
    virtual void forward(OverlayKey* key, cMessage* msg,
			 NodeHandle* nextHopNode);

    /**
     * Common API function: informs application about neighbors and own nodeID
     *
     * @param node new or lost neighbor
     * @param joined new or lost?
     */
    virtual void update(const NodeHandle& node, bool joined);

    /**
     * Common API function: produces a list of nodes that can be used as next
     * hops towards key
     *
     * @param key the destination key
     * @param num maximal number of nodes in answer
     * @param safe fraction of faulty nodes is not higher than in the overlay?
     */
    inline NodeVector* callLocalLookup(const OverlayKey& key, int num, bool safe)
    {
	return overlay->local_lookup(key, num, safe);
    };
    
    /**
     * Common API function: produces a list of neighbor nodes
     *
     * @param num maximal number of nodes in answer
     */
    inline NodeVector* callNeighborSet(int num)
    {
	return overlay->neighborSet(num);
    };
    
    /**
     * Common API function: produces a ordered list of neighbor nodes
     * on which replicas of key can be stored 
     *
     * @param key ID of the object
     * @param max_rank TODO
     */
    inline NodeVector* callReplicaSet(const OverlayKey& key, int max_rank)
    {
	return overlay->replicaSet(key, max_rank);
    };
    
    /**
     * Common API function: Information about the ranges of keys
     *
     * @param node the R-node of keys
     * @param rank ...
     * @param lKey ...
     * @param rKey ...
     */
    inline bool range(const NodeHandle& node, int rank,
		      const OverlayKey& lKey, const OverlayKey& rKey)
    {
	return overlay->range(node, rank, lKey, rKey);
    };

    /**
     * processes self-messages
     *
     * method to handle self-messages
     * should be overwritten in derived application if needed
     * @param msg self-message
     */
    virtual void handleTimerEvent(cMessage* msg);

    /**
     * method to handle non-commonAPI messages from the overlay
     *
     * @param msg message to handle
     */
    virtual void handleAppMessage(cMessage* msg);

    virtual void handleUpperMessage(cMessage* msg);

    /**
     * sends non-commonAPI message to the overlay
     *
     * @param msg message to send
     */
    void sendMessageToOverlay(cMessage* msg);

    NodeHandle thisNode; /**< NodeHandle to this node*/
    UnderlayConfigurator* underlayConfigurator; /**< pointer to
						   UnderlayConfigurator
						   in this node*/
    BootstrapOracle* bootstrapOracle; /**< pointer to BootstrapOracle
					 in this node*/

    GlobalStatistics* globalStatistics; /**< pointer to GlobalStatistics module
					 in this node*/

    BaseOverlay* overlay; /**< pointer to the overlay module of this node*/

    // parameters
    bool debugOutput; /**< debug output yes/no?*/
    bool onlyCommonAPIMessages; /**< process/send only commonAPI messages? */

    // statistics
    int numSent; /**< number of sent packets*/
    int bytesSent; /**< number of sent bytes*/
    int numReceived; /**< number of received packets*/
    int bytesReceived; /**< number of received bytes*/

public:

    /**
     * virtual destructor
     */
    virtual ~BaseApp();
};

#endif
