//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BootstrapOracle.cc
 * @author Markus Mauch, Robert Palmer
 */

#include <omnetpp.h>

#include <OverlayKey.h>

#include "BootstrapOracle.h"


Define_Module(BootstrapOracle);

std::ostream& operator<<(std::ostream& Stream, const bootstrapEntry entry)
{
    return Stream << "Address: " << entry.node->ip
		  << " Port: " << entry.node->port << " ModuleID: "
                  << entry.info->getModuleID() << " Bootstrapped: "
                  << (entry.info->isBootstrapped() ? "true" : "false");
}

void BootstrapOracle::initialize()
{
    maxNumberOfKeys = par("maxNumberOfKeys");
    keyProbability = par("keyProbability");
    //WATCH_MAP(peerSet);
    WATCH_VECTOR(keyList);
    WATCH(bootstrappedPeerSize);
    createKeyList(maxNumberOfKeys);
    bootstrappedPeerSize = 0;
}

void BootstrapOracle::handleMessage(cMessage* msg)
{
    error("this module doesn't handle messages, it runs only in initialize()");
}

const NodeHandle& BootstrapOracle::getBootstrapNode()
{
    if(bootstrappedPeerSize == 0)
        return NodeHandle::UNSPECIFIED_NODE;
    else {
	// return random TransportAddress in O(log n)
	PeerHashMap::iterator it = peerSet.end();
	bootstrapEntry tempEntry = {NULL, NULL};

	while(it == peerSet.end() || !it->second.info->isBootstrapped()) {
	    IPvXAddress randomAddr(intuniform(0,0xFFFFFFFF));

	    it = peerSet.find(randomAddr);

	    if (it == peerSet.end()) {
		it = peerSet.insert(std::make_pair(randomAddr,tempEntry)).first;
		peerSet.erase(it++);
	    }

	    if (it == peerSet.end())
		it = peerSet.begin();
	}

	if(dynamic_cast<NodeHandle*>(it->second.node))
	    return *dynamic_cast<NodeHandle*>(it->second.node);
	else
	    return NodeHandle::UNSPECIFIED_NODE;
    }
}

void BootstrapOracle::addPeer(const IPvXAddress& addr, PeerInfo* info)
{
    bootstrapEntry temp;
    temp.node = new TransportAddress(addr);
    temp.info = info;
    peerSet.insert(std::make_pair(addr, temp));
}

void BootstrapOracle::registerPeer(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if (it == peerSet.end())
        error("unable to bootstrap peer, peer is not in peer set");
    else {
        PeerInfo* info = it->second.info;

	if (!info->isBootstrapped())
	    bootstrappedPeerSize++;

        info->setBootstrapped();

        delete it->second.node;
        peerSet.erase(it);

        bootstrapEntry temp;
        temp.node = new TransportAddress(peer);
        temp.info = info;
        peerSet.insert(std::make_pair(temp.node->ip, temp));
        bootstrappedPeerSize++;
    }
}

void BootstrapOracle::registerPeer(const NodeHandle& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if (it == peerSet.end())
        error("unable to bootstrap peer, peer is not in peer set");
    else {
        PeerInfo* info = it->second.info;

	if (!info->isBootstrapped())
	    bootstrappedPeerSize++;

        info->setBootstrapped();
        delete it->second.node;
        peerSet.erase(it);

        bootstrapEntry temp;
        temp.node = new NodeHandle(peer);
        temp.info = info;
        peerSet.insert(std::make_pair(temp.node->ip, temp));
    }
}

void BootstrapOracle::removePeer(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if(it != peerSet.end()) {
	if (it->second.info->isBootstrapped())
	    bootstrappedPeerSize--;

        it->second.info->setBootstrapped(false);
    }
}

void BootstrapOracle::killPeer(const IPvXAddress& ip)
{
    PeerHashMap::iterator it = peerSet.find(ip);
    if(it != peerSet.end()) {
	if (it->second.info->isBootstrapped())
	    bootstrappedPeerSize--;

        it->second.info->setBootstrapped(false);
        delete it->second.node;
        delete it->second.info;
        peerSet.erase(it);
    }
}

void BootstrapOracle::createKeyList(uint size)
{
    for(uint i = 0; i < size; i++)
        keyList.push_back(OverlayKey::random());
}

BootstrapOracle::KeyList* BootstrapOracle::getKeyList(uint maximumKeys)
{
    if (maximumKeys > keyList.size()) {
	maximumKeys = keyList.size();
    }
    // copy keylist to temporary keylist
    KeyList tmpKeyList;
    tmpKeyList.clear();
    for ( uint i=0; i < keyList.size(); i++ )
        tmpKeyList.push_back(keyList[i]);

    KeyList* returnList = new KeyList;

    for ( uint i=0; i < ((float)maximumKeys * keyProbability); i++) {
        uint index = intuniform(0, tmpKeyList.size()-1);
        returnList->push_back(tmpKeyList[index]);
        tmpKeyList.erase(tmpKeyList.begin()+index);
    }

    return returnList;
}

const OverlayKey& BootstrapOracle::getRandomKeyListItem() const
{
    return keyList[intuniform(0,keyList.size()-1)];
}

PeerInfo* BootstrapOracle::getPeerInfo(const TransportAddress& peer)
{
    PeerHashMap::iterator it = peerSet.find(peer.ip);
    if(it == peerSet.end())
        return NULL;
    else
        return it->second.info;
}
