//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BootstrapOracle.h
 * @author Markus Mauch, Robert Palmer
 */

#ifndef __BOOTSTRAPORACLE_H__
#define __BOOTSTRAPORACLE_H__

class PeerInfo;

#include <map>
#include <vector>

#include <omnetpp.h>

#include <HashFunc.h>
#include <NodeHandle.h>
#include <OverlayKey.h>
#include <PeerInfo.h>


/**
 * Global modul that supports bootstrap process and key distribution
 *
 * @author Markus Mauch, Robert Palmer
 */
struct bootstrapEntry
{
    TransportAddress* node;
    PeerInfo* info;
    friend std::ostream& operator<<(std::ostream& Stream, const bootstrapEntry entry);
};

class BootstrapOracle : public cSimpleModule
{
public:

    typedef std::vector<OverlayKey> KeyList;

    /**
     * Adds new peers to the peer set.
     *
     * Called automatically by the underlay,
     * when new peers are created.
     *
     * @param peer node to add
     */
    //virtual void addPeer(TransportAddress* peer, PeerInfo* info);

    void addPeer(const IPvXAddress& addr, PeerInfo* info);

    /**
     * Removes a peer from the peer set.
     *
     * Called automatically by the underlay,
     * when peers are removed.
     *
     * @param peer node to remove
     */
    virtual void killPeer(const IPvXAddress& ip);

    /**
     * Returns a random transport address
     *
     * Returns a random transport address from the peer set if at least one peer has
     * been registered, and an empty transport address otherwise.
     *
     * @return bootstrap node
     */
    virtual const NodeHandle& getBootstrapNode();

    /**
     * Bootstraps peers in the peer set.
     *
     * @param peer node to register
     */
    virtual void registerPeer(const TransportAddress& peer);

    /**
     * Bootstraps peers in the peer set.
     *
     * @param peer node to register
     */
    virtual void registerPeer(const NodeHandle& peer);

    /**
     * Debootstraps peers in the peer set
     *
     * @param peer node to remove
     */
    virtual void removePeer(const TransportAddress& peer);

    /**
     * Returns a keylist
     *
     * @param maximumKeys maximum number of keys in new keylist
     * @return pointer to new keylist
     */
    virtual KeyList* getKeyList(uint maximumKeys);

    /**
     * Returns random key from list
     *
     * @return the key
     */
    virtual const OverlayKey& getRandomKeyListItem() const;

    /**
     * Returns peer information corresponding to a specified node
     *
     * Returns a random transport address from the peer set if at least one peer has
     * been registered, and an empty transport address otherwise.
     *
     * @return peer info
     */
    virtual PeerInfo* getPeerInfo(const TransportAddress& peer);

protected:

    /**
     * Init member function of module
     */
    virtual void initialize();

    /**
     * HandleMessage member function of module
     */
    virtual void handleMessage(cMessage* msg);

    /**
     * Member function to create keylist
     *
     * @param size size of new keylist
     */
    virtual void createKeyList(uint size);

    KeyList keyList; //!< the keylist
    uint bootstrappedPeerSize; //!< number of bootstrapped peers in the peer set

    typedef __gnu_cxx::hash_map<IPvXAddress, bootstrapEntry> PeerHashMap; //!< Set of nodes participating in the overlay
    PeerHashMap peerSet; //!< Set of nodes participating in the overlay

    // OMNeT++ parameters
    uint maxNumberOfKeys; //!< parameter used by createKeyList()
    double keyProbability; //!< probability of keys to be owned by nodes
};

#endif
