#ifndef __NODE_VECTOR_H
#define __NODE_VECTOR_H

#include <Comparator.h>
#include <NodeHandle.h>
#include <vector>

template <class T> class KeyExtractor;
template <class T, class T_key = KeyExtractor<T> > class BaseKeySortedVector;
typedef BaseKeySortedVector< NodeHandle > NodeVector;
typedef BaseKeySortedVector< std::pair<NodeHandle, NodeHandle> > NodePairVector;

/**
 * Class for extracting the relevant OverlayKey from a type used as template
 * parameter T_key for NodeVector<T, T_key> - Generic Version for unknown
 * types, returns unspecified keys
 *
 * @author Felix M. Palmen
 */
template <class T>
struct KeyExtractor
{
    static const OverlayKey& key(const T&)
    {
	return OverlayKey::UNSPECIFIED_KEY;
    };
};

/**
 * Class for extracting the relevant OverlayKey from a type used as template
 * parameter T_key for NodeVector<T, T_key> - Version for plain NodeHandle.
 *
 * @author Felix M. Palmen
 */
template <>
struct KeyExtractor<NodeHandle>
{
    static const OverlayKey& key(const NodeHandle& node)
    {
	return node.key;
    };
};

/**
 * Class for extracting the relevant OverlayKey from a type used as template
 * parameter T_key for NodeVector<T, T_key> - Version for a pair of
 * NodeHandles, first one is assumed to be relevant
 *
 * @author Felix M. Palmen
 */
template <>
struct KeyExtractor<std::pair<NodeHandle, NodeHandle> >
{
    static const OverlayKey& key(
	    const std::pair<NodeHandle, NodeHandle>& nodes)
    {
	return nodes.first.key;
    };
};


/**
 * A STL-vector that supports inserts sorted by an OverlayKey found somewhere
 * in the type
 * 
 * @author Sebastian Mies
 */
template <class T, class T_key>
class BaseKeySortedVector : public std::vector<T> {

private://fields: comparator

    const Comparator<OverlayKey>* comparator;
    uint16_t maxSize;

public://construction

    BaseKeySortedVector( uint16_t maxSize = 0,
	    const Comparator<OverlayKey>* comparator = NULL ) :
	std::vector<T>(),
	comparator(comparator),
	maxSize(maxSize) {};

    ~BaseKeySortedVector() {};

    typedef typename std::vector<T>::iterator iterator;
    typedef typename std::vector<T>::const_iterator const_iterator;

    static const T UNSPECIFIED_ELEMENT;

public://methods: sorted add support

    bool isAddable( const T& element ) const
    {
	return ( std::vector<T>::size() != maxSize ||
		(comparator && ( comparator->compare( T_key::key(element),
		    T_key::key(std::vector<T>::back()) ) < 0 )));
    };

    bool isFull() const
    {
	return (std::vector<T>::size() == maxSize);
    };

    bool isEmpty() const
    {
	return (std::vector<T>::size() == 0);
    };

    bool add( const T& element )
    {
        // check if handle is addable
        if ( isAddable(element) )
	{ // yes ->

	    // add handle to the appropriate position
	    if ( ( std::vector<T>::size() != 0 ) && comparator )
	    {
		for ( iterator i = std::vector<T>::begin();
			i != std::vector<T>::end(); i++ )
		    if ( comparator->compare(T_key::key(element),
				T_key::key(*i)) < 0 )
		    {
			std::vector<T>::insert( i, element );
			break;
		    }
	    }
	    else
	    {
		push_back(element);
	    }

	    // adjust size
	    if ((maxSize != 0) && (std::vector<T>::size() >= maxSize))
	    {
		std::vector<T>::resize(maxSize);
	    }

	    return true;
	}
	else
	{
	    return false;
	}

    };

    const T& find( const OverlayKey& key ) const
    {
	for ( const_iterator i = std::vector<T>::begin();
		i != std::vector<T>::end(); i++)
	{
	    if (T_key::key(*i) == key) return *i;
	}
	return UNSPECIFIED_ELEMENT;
    };
};

template <class T, class T_key>
const T BaseKeySortedVector<T, T_key>::UNSPECIFIED_ELEMENT;

#endif
