//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file TransportAddress.h
 * @author Markus Mauch
 * @author Sebastian Mies
 */

#ifndef __TRANSPORTADDRESS_H_
#define __TRANSPORTADDRESS_H_

//class TransportAddress;

#include <stdint.h>
#include <omnetpp.h>
#include <IPvXAddress.h>

#include <ext/hash_set>
#include <ext/hash_map>

using namespace __gnu_cxx;
using namespace std;

/**
 * This class implements a common transport address.<br>
 *
 * It covers the complete node information, like IP-Address,
 * and port. The information can be sparse, so parts can be
 * omited by setting the property to an unspecified value.
 *
 * @author Markus Mauch
 * @author Sebastian Mies
 */

class TransportAddress
{
public:
    class hashFcn
    {
    public:
        size_t operator()( const TransportAddress& h1 ) const
        {
            return h1.hash();
        }
    };

public://collection typedefs

    typedef hash_set<TransportAddress, hashFcn> Set;

public://fields

    IPvXAddress ip;
    int port;

public://construction

    /**
     * Constructs a unspecified TransportAddress
     */
    TransportAddress( );

    /**
     * Standard destructor
     */
    virtual ~TransportAddress( ) {};

    /**
     * Copy constructor.
     *
     * @param handle The TransportAddress to copy
     */
    TransportAddress( const TransportAddress& handle );

    /**
     * Complete constructor.
     *
     * @param ip The IPvXAddress
     * @param port The UDP-Port
     */
    TransportAddress(const IPvXAddress& ip, int port = -1);

public://static fields

    static const TransportAddress UNSPECIFIED_NODE;

public://methods: delegates to OverlayKey and IPvXAddress

    bool operator==(const TransportAddress& rhs) const;
    bool operator!=(const TransportAddress& rhs) const;
    bool operator< (const TransportAddress& rhs) const;
    bool operator> (const TransportAddress& rhs) const;
    bool operator<=(const TransportAddress& rhs) const;
    bool operator>=(const TransportAddress& rhs) const;

public://methods: operators

    TransportAddress& operator=(const TransportAddress& rhs);

public://methods: setters and getters

    void setAddress( const IPvXAddress& ip, int port = -1 );
    void setPort( int port );

    const IPvXAddress& getAddress() const;
    int getPort() const;

    bool isUnspecified() const;

public://methods: hashing

    size_t hash() const;

public://methods: c++ streaming

    friend std::ostream& operator<<(std::ostream& os, const TransportAddress& n);

private://methods:

    void assertUnspecified( const TransportAddress& handle ) const;

public:

    virtual void netPack(cCommBuffer *b);
    virtual void netUnpack(cCommBuffer *b);
};

inline void doPacking(cCommBuffer *b, TransportAddress& obj) {obj.netPack(b);}
inline void doUnpacking(cCommBuffer *b, TransportAddress& obj) {obj.netUnpack(b);}

#endif
