//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BrooseBucket.h
 * @author Jochen Schenk
 */

#ifndef __BROOSEBUCKET_H_
#define __BROOSEBUCKET_H_

#include <omnetpp.h>
#include "BrooseHandle.h"
#include <map>
#include <OverlayKey.h>


/**
 * Broose bucket module
 *
 * This modul contains the bucket of the Broose implementation.
 *
 * @author Jochen Schenk
 * @see Broose
 */
class BrooseBucket : public cSimpleModule
{
public:
    //omnet
    virtual int numInitStages() const
    {
        return 6;
    }
    virtual void initialize(int stage);
    virtual void handleMessage(cMessage* msg);

    // bucket functions
    /**
     * adds a broose node handle to the bucket
     *
     * @param node broose handle which will be added
     */
    virtual void add
        (BrooseHandle node);

    /**
     * removes a broose node handle from the bucket
     *
     * @param node broose handle which will be removed
     */
    virtual void remove
        (BrooseHandle node);

    /**
     * returns a specific broose handle
     *
     * @param pos position of the broose handle in the bucket
     */
    virtual const BrooseHandle& get
        (uint pos = 0);

    /**
     * returns distance of a specific broose handle
     *
     * @param pos position of the broose handle in the bucket
     */
    virtual const OverlayKey& getDist(uint pos = 0);

    /**
     * initializes a bucket
     *
     * @param shiftingBits specifies the kind of the bucket
     * @param prefix in case of a R bucket specifies the prefix
     * @param node broose handle to the node which owns this bucket
     * @param size maximal size of the bucket
     */
    virtual void initializeBucket(int shiftingBits, uint prefix, BrooseHandle node, int size);

    /**
     * returns number of current entries
     *
     * @return number of current entries 
     */
    virtual uint getSize();

    /**
     * returns number of maximal entries
     *
     * @return number of maximal entries 
     */
    virtual uint getMaxSize();

    /**
     * checks if the bucket is empty
     *
     * @return true if the bucket is empty
     */
    virtual bool isEmpty();

    /**
     * removes all entries from the bucket
     */
    virtual void clear();

    /**
     * returns the closest node to a specific key
     *
     * @param destKey key to compare
     * @return broose handle of closest node
     */
    virtual BrooseHandle getClosestNode(const OverlayKey& destKey);

    /**
     * return the longest prefix of all entries
     *
     * @return the longest prefix
     */
    virtual int longestPrefix(void);

    /**
     * checks if the key close to the owner's id
     *
     * @param key key to check 
     * @return true if node is close
     */
    virtual bool keyInRange(const OverlayKey& key);

    /**
     * checks if a node is already in the bucket
     *
     * @param node broose handle to check
     * @return true if node is already in the bucket
     */
    virtual bool nodeInBucket(const BrooseHandle& node);

    /**
     * returns the number of failed responses to a specific broose handle
     *
     * @param node broose handle to check
     * @return number of failed responses
     */
    virtual int getFailedResponses (const BrooseHandle& node);

    /**
     * increase the number of failed responses to a specific broose handle
     *
     * @param node broose handle which counter will be increased
     */
    virtual void increaseFailedResponses (const BrooseHandle& node);

    /**
     * resets the counter of failed responses to a specific broose handle
     *
     * @param node broose handle which counter will be reset
     */
    virtual void resetFailedResponses (const BrooseHandle& node);

    /**
     * sets the round trip time to a specific broose handle
     *
     * @param node broose handle to which the rtt will be stored
     * @param rpcRTT rtt to the specific node
     */
    virtual void setRTT(BrooseHandle node, simtime_t rpcRTT);

    /**
     * returns the round trip time to a specific broose handle
     *
     * @param node broose handle to which the rtt will be retrieved
     * @return rtt to the specific node
     */
    virtual simtime_t getRTT(BrooseHandle node);

    /**
     * updates the timestamp of a specific node
     *
     * @param node broose handle to which the timestamp will be updated
     * @param lastSeen timestamp of the last message
     */
    virtual void setLastSeen(BrooseHandle node, simtime_t lastSeen);

    /**
     * returns the timestamp of a specific node
     *
     * @param node broose handle to which the timestamp will be retrieved
     * @return the retrieved timestamp
     */
    virtual simtime_t getLastSeen(BrooseHandle node);

    /**
     * displays the content of the bucket on standard out
     *
     * @param maxEntries the maximal number of entries which will be printed
     */
    void output(int maxEntries = 0);

    /**
     * displays a key in binary form
     *
     * @param key key which is displayed
     */
    void binaryOutput(const OverlayKey& key);

protected:
    // parameter
    std::map<const OverlayKey, BrooseHandle> bucket; /**< data structure representing the bucket */
    std::map<const OverlayKey, BrooseHandle>::iterator bucketIter; /**< iterator to navigate through the bucket */

    unsigned int maxSize; /**< maximal size of the bucket */
    BrooseHandle thisNode; /**< node handle to the node managing this bucket */
    OverlayKey key; /**< the node's key shifted to fit the bucket and used to measure distance to other keys */
    int bits; /**< number of bits the node's key has to be shifted to fit the apropriate bucket */


    // test and debug functions
    void test();
};
#endif
