//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Pastry.cc
 * @author Felix Palmen
 */

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>
#include <RpcMacros.h>
#include <InitStages.h>

#include "Pastry.h"


Define_Module(Pastry);

void Pastry::initializeOverlay(int stage)
{
    if ( stage != MIN_STAGE_OVERLAY )
        return;

    bitsPerDigit = par("bitsPerDigit");
    numberOfLeaves = par("numberOfLeaves");
    numberOfNeighbors = par("numberOfNeighbors");
    joinTimeoutAmount = par("joinTimeout");
    readyWaitAmount = par("readyWait");
    secondStageWaitAmount = par("secondStageWait");
    pingCacheExpireTime = par("pingCacheExpireTime");
    repairTimeout = par("repairTimeout");
    enableNewLeafs = par("enableNewLeafs");
    optimizeLookup = par("optimizeLookup");
    optimisticForward = par("optimisticForward");
    avoidDuplicates = par("avoidDuplicates");
    ringCheckInterval = par("ringCheckInterval");

    if (numberOfLeaves % 2)
    {
	EV << "Pastry: Warning: numberOfLeaves must be even - adding 1.";
	numberOfLeaves++;
    }

    routingTable = check_and_cast<PastryRoutingTable*>
	(parentModule()->submodule("pastryRoutingTable"));
    leafSet = check_and_cast<PastryLeafSet*>
	(parentModule()->submodule("pastryLeafSet"));
    neighborhoodSet = check_and_cast<PastryNeighborhoodSet*>
	(parentModule()->submodule("pastryNeighborhoodSet"));

    joinTimeout = new cMessage("joinTimeout");
    readyWait = new cMessage("readyWait");
    secondStageWait = new cMessage("secondStageWait");
    ringCheck = new cMessage("ringCheck");

    stateCache = NULL;
    
    // initialize statistics
    joins = 0;
    joinTries = 0;
    joinSeen = 0;
    joinReceived = 0;
    joinSent = 0;
    stateSent = 0;
    stateReceived = 0;
    repairReqSent = 0;
    repairReqReceived = 0;
    stateReqSent = 0;
    stateReqReceived = 0;

    joinBytesSeen = 0;
    joinBytesReceived = 0;
    joinBytesSent = 0;
    stateBytesSent = 0;
    stateBytesReceived = 0;
    repairReqBytesSent = 0;
    repairReqBytesReceived = 0;
    stateReqBytesSent = 0;
    stateReqBytesReceived = 0;

    WATCH(joins);
    WATCH(joinTries);
    WATCH(joinSeen);
    WATCH(joinBytesSeen);
    WATCH(joinReceived);
    WATCH(joinBytesReceived);
    WATCH(joinSent);
    WATCH(joinBytesSent);
    WATCH(stateSent);
    WATCH(stateBytesSent);
    WATCH(stateReceived);
    WATCH(stateBytesReceived);
    WATCH(repairReqSent);
    WATCH(repairReqBytesSent);
    WATCH(repairReqReceived);
    WATCH(repairReqBytesReceived);
    WATCH(stateReqSent);
    WATCH(stateReqBytesSent);
    WATCH(stateReqReceived);
    WATCH(stateReqBytesReceived);
    WATCH(lastStateChange);
    WATCH(proxCache);
    WATCH_VECTOR(stateReceivedProximities);


    changeState(INIT);

    if (bootstrapNode.isUnspecified())
    {
	changeState(READY);
    }
    else
    {
	changeState(JOIN);
    }
}

void Pastry::purgeVectors(void)
{
    PastryStateMessage* stateMsg;

    // purge pending state messages
    if (!stReceived.empty())
    {   
	for (std::vector<PastryStateMessage*>::iterator it =
		stReceived.begin(); it != stReceived.end(); it++)
	{
	    // check whether one of the pointers is a duplicate of stateCache
	    if (*it == stateCache) stateCache = NULL;
	    if (debugOutput)
	    {
		EV << "Pastry::purgeVectors() deleting STATE message "
		    << static_cast<void*>(*it) << endl;
	    }
	    delete *it;
	}
	stReceived.clear();
    }

    // purge notify list:
    notifyList.clear();

    // purge Ping Cache:
    pingCache.clear();

    // purge proximity cache for received state messages:
    stateReceivedProximities.clear();

    // purge Queue for messages to be forwarded in recursive mode:
    for (std::map<TransportAddress, BaseRouteMessage*>::iterator i =
	    recFwdQueue.begin(); i != recFwdQueue.end(); i++)
    {
	delete i->second;
    }
    recFwdQueue.clear();

    // purge Queue for processing multiple STATE messages:
    while (! stateCacheQueue.empty())
    {
	stateMsg = (PastryStateMessage*) stateCacheQueue.pop();
	delete stateMsg;
    }
    
    // delete cached state message:
    if (stateCache)
    {
	if (debugOutput)
	{
	    EV << "Pastry::purgeVectors() deleting STATE message "
		<< static_cast<void*>(stateCache) << endl;
	}
	delete stateCache;
	stateCache = NULL;
    }
}

void Pastry::newLeafs(void)
{
    if (! enableNewLeafs) return;

    PastryNewLeafsMessage* msg = leafSet->getNewLeafsMessage();
    if (msg)
    {
	send(msg, "to_app");
	EV << "Pastry: newLeafs() called." << endl;
    }
}

void Pastry::reJoin(void)
{
    if (ringCheck->isScheduled()) cancelEvent(ringCheck);
    if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);
    if (readyWait->isScheduled()) cancelEvent(readyWait);
    bootstrapOracle->removePeer(thisNode);
    purgeVectors();
    bootstrapNode = bootstrapOracle->getBootstrapNode();
    routingTable->initializeTable(bitsPerDigit, repairTimeout, thisNode);
    leafSet->initializeSet(numberOfLeaves, repairTimeout, thisNode);
    neighborhoodSet->initializeSet(numberOfNeighbors, thisNode);
    updateTooltip();
    lastStateChange = simTime();
    changeState(JOIN);
}

void Pastry::changeState(int toState)
{
    if (ringCheck->isScheduled()) cancelEvent(ringCheck);

    switch (toState)
    {
	case INIT:
	    
	    state = INIT;

	    if (!thisNode.key.isUnspecified())
		bootstrapOracle->removePeer(thisNode);

	    if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);
	    if (readyWait->isScheduled()) cancelEvent(readyWait);

	    purgeVectors();
	    
	    thisNode.key = OverlayKey::random();
	    callUpdate(thisNode, true);
	    bootstrapNode = bootstrapOracle->getBootstrapNode();
	    
	    routingTable->initializeTable(bitsPerDigit, repairTimeout,
		    thisNode);
	    leafSet->initializeSet(numberOfLeaves, repairTimeout, thisNode);
	    neighborhoodSet->initializeSet(numberOfNeighbors, thisNode);

	    updateTooltip();
	    lastStateChange = simTime();
    
	    parentModule()->parentModule()->bubble("entering INIT state");

	    break;

	case JOIN:

	    state = JOIN;
	    joinHopCount = 0;

	    {
		PastryJoinMessage* msg = new PastryJoinMessage("JOIN-Request");
		msg->setPastryMsgType(PASTRY_MSG_JOIN);
		msg->setSignaling(true);
		msg->setJoinHopCount(1);
		msg->setSendStateTo(thisNode);
		msg->setLength(PASTRYJOIN_L(msg));
		RECORD_STATS(joinSent++; joinBytesSent += msg->byteLength());
	        sendToKey(thisNode.key, msg, 0, bootstrapNode);
	    }
	    
	    cancelEvent(joinTimeout);
	    scheduleAt(simTime()+joinTimeoutAmount, joinTimeout);

	    updateTooltip();
	    parentModule()->parentModule()->bubble("entering JOIN state");

	    RECORD_STATS(joinTries++);

	    break;
	
	case READY:

	    state = READY;
	    if (ringCheckInterval > 0)
	    {
		scheduleAt(simTime()+ringCheckInterval, ringCheck);
	    }

	    bootstrapOracle->registerPeer(thisNode);

	    // if we are the first node in the network, there's nothing else
	    // to do
	    if (bootstrapNode.isUnspecified())
	    {
		RECORD_STATS(joinTries++);
		RECORD_STATS(joins++);
		setReadyIcon(true);
		return;
	    }

	    // determine list of all known nodes as notifyList
	    notifyList.clear();
	    leafSet->dumpToVector(notifyList);
	    routingTable->dumpToVector(notifyList);
	    sort(notifyList.begin(), notifyList.end());
	    notifyList.erase(unique(notifyList.begin(), notifyList.end()),
		    notifyList.end());
	
	    // send "update" state message to all nodes who sent us their state
	    // during INIT, remove these from notifyList so they don't get our
	    // state twice
	    std::vector<TransportAddress>::iterator nListPos;
	    if (!stReceived.empty())
	    {
		for (std::vector<PastryStateMessage*>::iterator it =
			stReceived.begin(); it != stReceived.end(); it++)
		{
		    sendStateTables((*it)->getSender(), PASTRY_STATE_UPDATE,
			    (*it)->getTimestamp());
		    nListPos = find(notifyList.begin(), notifyList.end(),
			    (*it)->getSender());
		    if (nListPos != notifyList.end())
		    {
			notifyList.erase(nListPos);
		    }
		    if (debugOutput)
		    {
			EV << "Pastry::changeState() deleting STATE message "
			    << static_cast<void*>(*it) << endl;
		    }
		    delete *it;
		}
		stReceived.clear();
	    }

	    // send a normal STATE message to all remaining known nodes
	    for (std::vector<TransportAddress>::iterator it =
		    notifyList.begin(); it != notifyList.end(); it++)
	    {
		if (*it != thisNode) sendStateTables(*it);
	    }

	    updateTooltip();
	    parentModule()->parentModule()->bubble("entering READY state");

	    // schedule second stage
	    cancelEvent(secondStageWait);
	    scheduleAt(simTime() + secondStageWaitAmount, secondStageWait);

	    RECORD_STATS(joins++);

	    break;
    }
    setReadyIcon(state == READY);
}

void Pastry::handleTimerEvent(cMessage* msg)
{

    if (msg->isName("joinTimeout"))
    {
	EV << "Pastry: join timeout expired, restarting..." << endl;
	changeState(INIT);
	changeState(JOIN);
    }
    else if (msg->isName("readyWait"))
    {
	EV << "Pastry: timeout waiting for missing state messages in JOIN "
	    "state, restarting..." << endl;
	changeState(INIT);
	changeState(JOIN);
    }
    else if (msg->isName("secondStageWait"))
    {
	EV << "Pastry: sending STATE requests to all nodes in second stage "
	    "of initialization." << endl;
	doSecondStage();
    }
    else if (msg->isName("ringCheck"))
    {
	if (state == READY)
	{
	    // ping direct neighbors on the ring:
	    const NodeHandle& pred = leafSet->getPredecessor();
	    const NodeHandle& succ = leafSet->getSuccessor();
	    if (! pred.isUnspecified())
	    {
		sendRpcMessage(pred, new PingCall("PING ring check"),
			NULL, OverlayKey::UNSPECIFIED_KEY, -1, 5.0, 2);
	    }
	    if (! succ.isUnspecified())
	    {
		sendRpcMessage(succ, new PingCall("PING ring check"),
			NULL, OverlayKey::UNSPECIFIED_KEY, -1, 5.0, 2);
	    }
	}
	scheduleAt(simTime()+ringCheckInterval, ringCheck);
    }
}

void Pastry::handleUDPMessage(BaseOverlayMessage* msg)
{
    PastryMessage* pastryMsg = check_and_cast<PastryMessage*>(msg);
    uint type = pastryMsg->getPastryMsgType();

    if (debugOutput)
    {
	EV << "Pastry: incoming message of type ";
	switch(type)
	{
	    case PASTRY_MSG_STD: EV << "PASTRY_MSG_STD"; break;
	    case PASTRY_MSG_JOIN: EV << "PASTRY_MSG_JOIN"; break;
	    case PASTRY_MSG_STATE: EV << "PASTRY_MSG_STATE"; break;
	    case PASTRY_MSG_RREQ: EV << "PASTRY_MSG_RREQ"; break;
	    case PASTRY_MSG_SREQ: EV << "PASTRY_MSG_SREQ"; break;
	    default: EV << "UNKNOWN (" << type <<")"; break;
	}
	EV << endl;
    }
    
    switch (type)
    {
	case PASTRY_MSG_STD:
	    opp_error("Pastry received PastryMessage of unknown type!");
	    break;
	    
	case PASTRY_MSG_JOIN:
	    {
		PastryJoinMessage* jmsg =
		    check_and_cast<PastryJoinMessage*>(pastryMsg);
		RECORD_STATS(joinReceived++; joinBytesReceived +=
			jmsg->byteLength());
		if (! iterativeLookup)
		{
		    // in iterative lookup mode, this is handled by findNode()
		    if (state == READY)
			sendStateTables(jmsg->getSendStateTo(),
				PASTRY_STATE_JOIN, jmsg->getJoinHopCount(),
				true);
		    else
			EV << "Pastry: received join message before reaching "
			    << "READY state, dropping message!" << endl;
		}
		delete jmsg;
	    }
	    break;

	case PASTRY_MSG_RREQ:
	    {
		PastryRepairRequestMessage* rrmsg =
		    check_and_cast<PastryRepairRequestMessage*>(pastryMsg);
		RECORD_STATS(repairReqReceived++; repairReqBytesReceived +=
			    rrmsg->byteLength());
		if (state == READY)
		    sendStateTables(rrmsg->getSendStateTo(),
			    PASTRY_STATE_REPAIR);
		else
		    EV << "Pastry: received repair request before reaching "
			<< "READY state, dropping message!" << endl;
		delete rrmsg;
	    }
	    break;
			
	case PASTRY_MSG_SREQ:
	    {
		PastryStateRequestMessage* srmsg =
		    check_and_cast<PastryStateRequestMessage*>(pastryMsg);
		RECORD_STATS(stateReqReceived++; stateReqBytesReceived +=
			    srmsg->byteLength());
		if (state == READY)
		    sendStateTables(srmsg->getSendStateTo());
		else
		    EV << "Pastry: received state request before reaching "
			<< "READY state, dropping message!" << endl;
		delete srmsg;
	    }
	    break;

	case PASTRY_MSG_STATE:
	    {
		PastryStateMessage* stateMsg =
		    check_and_cast<PastryStateMessage*>(msg);
		RECORD_STATS(stateReceived++; stateBytesReceived +=
			    stateMsg->byteLength());
		handleStateMessage(stateMsg);
	    }
	    break;
    }
}

void Pastry::handleStateMessage(PastryStateMessage* msg)
{
    if (debugOutput)
    {
	EV << "Pastry::handleStateMessage() new STATE message to process "
	    << static_cast<void*>(msg) << " in state " <<
	    ((state == READY)?"READY":((state == JOIN)?"JOIN":"INIT"))
	    << endl;
	if (state == JOIN)
	{
	    EV << " ***   own joinHopCount:  " << joinHopCount << endl;
	    EV << " ***   already received:  " << stReceived.size() << endl;
	    EV << " ***   last-hop flag:     " << msg->getLastHop() << endl;
	    EV << " ***   msg joinHopCount:  "
		<< msg->getJoinHopCount() << endl;
	}   
    }
    if (state == INIT)
    {
	EV << "Pastry: can't handle state messages until at least reaching "
	    "JOIN state." << endl;
	if (debugOutput)
	{
	    EV << "Pastry::handleStateMessage() deleting STATE message "
		<< static_cast<void*>(msg) << endl;
	}
	delete msg;
	return;
    }
    
    // in JOIN state, store all received state Messages, need them later:
    if (state == JOIN)
    {
	if (joinHopCount && stReceived.size() == joinHopCount)
	{
	    EV << "Pastry: Warning: dropping state message received after "
		"all needed state messages were collected in JOIN state."
		<< endl;
	    if (debugOutput)
	    {
		EV << "Pastry::handleStateMessage() deleting STATE message "
		    << static_cast<void*>(msg) << endl;
	    }
	    delete msg;
	    return;
	}

	stReceived.push_back(msg);
	    
	if (msg->getLastHop())
	{
	    if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);

	    if (joinHopCount)
	    {
		EV << "Pastry: Error: received a second `last' state message! "
		    "Restarting ..." << endl;
		changeState(INIT);
		changeState(JOIN);
		return;
	    }
		
	    joinHopCount = msg->getJoinHopCount();
	    if (stReceived.size() < joinHopCount)
	    {
		// some states still missing:
		cancelEvent(readyWait);
	    	scheduleAt(simTime()+readyWaitAmount, readyWait);
	    }
	}

	if (joinHopCount)
	{
	    if (stReceived.size() > joinHopCount)
	    {
		EV << "Pastry: Error: too many state messages received in "
		    "JOIN state! Restarting ..." << endl;
		changeState(INIT);
		changeState(JOIN);
		return;
	    }
	    if (stReceived.size() == joinHopCount)
	    {
		// all state messages are here, sort by hopcount:
		sort(stReceived.begin(), stReceived.end(),
			stateMsgIsSmaller);

		// start pinging the nodes found in the first state message:
		stateReceivedProximities.clear();
		stateCache = stReceived.front();
		EV << "Pastry: have all STATE messages, now pinging nodes."
		    << endl;
		pingNodes();

		// cancel timeout:
		if (readyWait->isScheduled()) cancelEvent(readyWait);
	    }
	}
	return;
    }

    // in READY state, only ping nodes to get proximity metric:
    if (!stateCache)
    {
	// no state message is processed right now, start immediately:
	stateCache = msg;
	pingNodes();
    }
    else
    {
	// enqueue message for later processing:
	stateCacheQueue.insert(msg);
    }
}

void Pastry::forwardMessageRecursive(const TransportAddress& dest,
	BaseRouteMessage* msg)
{
    PastryMessage* pmsg = dynamic_cast<PastryMessage*>(
	    msg->encapsulatedMsg());
    if ( pmsg && (dest != thisNode) )
    {
	if (pmsg->getPastryMsgType() == PASTRY_MSG_JOIN)
	{
	    PastryJoinMessage* jmsg = check_and_cast<PastryJoinMessage*>(pmsg);
	    RECORD_STATS(joinSeen++; joinBytesSeen += jmsg->byteLength());
	    sendStateTables(jmsg->getSendStateTo(), PASTRY_STATE_JOIN,
		    jmsg->getJoinHopCount(), false);
	    jmsg->setJoinHopCount(jmsg->getJoinHopCount() + 1);
	}
    }

    if (recFwdQueue.find(dest) != recFwdQueue.end())
    {
	// rare case, other message for same next hop is pending:
	// send the message and hope it isn't lost.
	sendMessageToUDP(dest, msg);
	return;
    }

    if (optimisticForward)
    {
	// forward now:
	sendMessageToUDP(dest, msg);
	if (! avoidDuplicates)
	{
	    // and keep copy for possible retry:
	    recFwdQueue[dest] = static_cast<BaseRouteMessage*>(msg->dup());
	}
    }
    else
    {
	// keep message in queue (forward later)
	recFwdQueue[dest] = msg;
    }
    sendRpcMessage(dest, new PingCall("PING next hop"),
	    NULL, OverlayKey::UNSPECIFIED_KEY, -1, 5.0, 2);
}

void Pastry::handleRpc(BaseCallMessage* msg)
{
    RPC_SWITCH_START(msg)
	RPC_DELEGATE(Ping, rpcPing);
    RPC_SWITCH_END()
}

void Pastry::rpcPing(PingCall* call)
{
    sendRpcResponse(call, new PingResponse("PING response"));
}

void Pastry::doSecondStage(void)
{
    // "second stage" for locality:
    notifyList.clear();
    routingTable->dumpToVector(notifyList);
    neighborhoodSet->dumpToVector(notifyList);
    sort(notifyList.begin(), notifyList.end());
    notifyList.erase(unique(notifyList.begin(), notifyList.end()),
	    notifyList.end());
    for (std::vector<TransportAddress>::iterator it = notifyList.begin();
	    it != notifyList.end(); it++)
    {
	if (*it == thisNode) continue;
	EV << "second stage: requesting state from " << *it << endl;
	PastryStateRequestMessage* msg =
		new PastryStateRequestMessage("SREQ");
	msg->setSignaling(true);
	msg->setPastryMsgType(PASTRY_MSG_SREQ);
	msg->setSendStateTo(thisNode);
	msg->setLength(PASTRYSREQ_L(msg));
	RECORD_STATS(stateReqSent++; stateReqBytesSent +=
		msg->byteLength());
	sendMessageToUDP(*it, msg);
    }
}

void Pastry::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
	simtime_t rtt)
{
    const NodeHandle& src = msg->getSrcNode();
    const NodeHandle* node;
    int i;
    int n;
    PastryPingCacheEntry pce;
    std::map<TransportAddress, BaseRouteMessage*>::iterator pos;

    RPC_SWITCH_START(msg)
	RPC_ON_RESPONSE(Ping)
	{
	    if (stateCache)
	    {
		// insert into pingCache
		pce.inserted = simTime();
		pce.rtt = rtt;
		pingCache[src.ip] = pce;

		// look for node in routing table of processed state message:
		n = stateCache->getRoutingTableArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getRoutingTable(i));
		    if ((!node->isUnspecified()) && (*node == src))
			*(proxCache.pr_rt.begin()+i) = rtt;
		}

		// look for node in leaf set of processed state message:
		n = stateCache->getLeafSetArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getLeafSet(i));
		    if ((!node->isUnspecified()) && (*node == src))
			*(proxCache.pr_ls.begin()+i) = rtt;
		}

		// look for node in neighb. set of processed state message:
		n = stateCache->getNeighborhoodSetArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getNeighborhoodSet(i));
		    if ((!node->isUnspecified()) && (*node == src))
			*(proxCache.pr_ns.begin()+i) = rtt;
		}

		checkProxCache();
	    }

	    if ((pos = recFwdQueue.find(src)) != recFwdQueue.end())
	    {
		// insert into pingCache
		pce.inserted = simTime();
		pce.rtt = rtt;
		pingCache[src.ip] = pce;

		// send message
		if (!optimisticForward)
		{
		    sendMessageToUDP(pos->first, pos->second);
		}
		else if (!avoidDuplicates)
		{
		    delete pos->second;
		}
		recFwdQueue.erase(pos);
	    }

	    break;
	}
    RPC_SWITCH_END()
}

void Pastry::pingNodes(void)
{
    int i;
    int rt_size;
    int ls_size;
    int ns_size;
    const NodeHandle* node;
    std::vector<simtime_t>::iterator proxPos;
    PingCall* call;
    PastryPingCache::iterator it;

    rt_size = stateCache->getRoutingTableArraySize();
    proxCache.pr_rt.clear();
    proxCache.pr_rt.resize(rt_size, PASTRY_PROX_PENDING);

    ls_size = stateCache->getLeafSetArraySize();
    proxCache.pr_ls.clear();
    proxCache.pr_ls.resize(ls_size, PASTRY_PROX_PENDING);

    ns_size = stateCache->getNeighborhoodSetArraySize();
    proxCache.pr_ns.clear();
    proxCache.pr_ns.resize(ns_size, PASTRY_PROX_PENDING);

    std::vector<OverlayKey>* pinged = new std::vector<OverlayKey>;
    pinged->reserve(rt_size + ls_size + ns_size);
    std::vector<OverlayKey>::iterator pingedPos;

    for (i = 0; i < rt_size + ls_size + ns_size; i++)
    {
	if (i < rt_size)
	{
	    node = &(stateCache->getRoutingTable(i));
	    proxPos = proxCache.pr_rt.begin() + i;
	}
	else if ( i < (rt_size + ls_size) )
	{
	    node = &(stateCache->getLeafSet(i - rt_size));
	    proxPos = proxCache.pr_ls.begin() + (i - rt_size);
	}
	else
	{
	    node = &(stateCache->getNeighborhoodSet(i - rt_size - ls_size));
	    proxPos = proxCache.pr_ns.begin() + (i - rt_size - ls_size);
	}

	// proximity is undefined for unspecified nodes:
	if (node->isUnspecified())
	{
	    *proxPos = PASTRY_PROX_UNDEF;
	}

	// and 0 for own node:
	else if (*node == thisNode)
	{
	    *proxPos = 0;
	}

	// else determine real value:
	else
	{
	    if ((it = pingCache.find(node->ip)) != pingCache.end())
	    {
		if ((simTime() - it->second.inserted) < pingCacheExpireTime)
		{
		    *proxPos = it->second.rtt;
		    if (debugOutput)
		    {
			EV << "Pastry: pingCache HIT" << endl;
		    }
		    continue;
		}
		else
		{
		    if (debugOutput)
		    {
			EV << "Pastry: pingCache OUTDATED" << endl;
		    }
		    pingCache.erase(it);
		}
	    }
	    else if (debugOutput)
	    {
		EV << "Pastry: pingCache MISS" << endl;
	    }
	    if ((pingedPos = find(pinged->begin(), pinged->end(), node->key))
		    == pinged->end())
	    {
		pinged->push_back(node->key);
		call = new PingCall("PING received state");
		sendRpcMessage(*node, call, NULL,
			OverlayKey::UNSPECIFIED_KEY, -1, 5.0, 2);
	    }
	}
    }
    delete pinged;
    checkProxCache();
}

void Pastry::checkProxCache(void)
{
    simtime_t now = simTime();

    // no cached STATE message?
    if (!stateCache) return;

    // no entries in proxCache?
    if (proxCache.pr_rt.empty()) return;
    if (proxCache.pr_ls.empty()) return;
    if (proxCache.pr_ns.empty()) return;

    // some entries not yet determined?
    if (find(proxCache.pr_rt.begin(), proxCache.pr_rt.end(),
		PASTRY_PROX_PENDING) != proxCache.pr_rt.end()
       ) return;
    if (find(proxCache.pr_ls.begin(), proxCache.pr_ls.end(),
		PASTRY_PROX_PENDING) != proxCache.pr_ls.end()
       ) return;
    if (find(proxCache.pr_ns.begin(), proxCache.pr_ns.end(),
		PASTRY_PROX_PENDING) != proxCache.pr_ns.end()
       ) return;

    if (state == JOIN)
    {
	stateReceivedProximities.push_back(proxCache);
	proxCache.pr_rt.clear();
	proxCache.pr_ls.clear();
	proxCache.pr_ns.clear();

	// collected proximities for all STATE messages?
	if (stateReceivedProximities.size() == stReceived.size())
	{
	    stateCache = NULL;
	    if (debugOutput)
	    {
		EV << "Pastry: [JOIN] starting to build own state from "
		    << stReceived.size() << " received state messages..."
		    << endl;
		EV << "Pastry: [JOIN] initializing NeighborhoodSet from "
		    << stReceived.front()->getJoinHopCount() << ". hop"
		    << endl;
	    }
	    if (!neighborhoodSet->mergeState(stReceived.front(),
			stateReceivedProximities.front()))
	    {
		EV << "Pastry: Error initializing own neighborhoodSet while "
		    "joining! Restarting ..." << endl;
		changeState(INIT);
		changeState(JOIN);
		return;
	    }
	    if (debugOutput)
	    {
		EV << "Pastry: [JOIN] initializing LeafSet from "
		    << stReceived.back()->getJoinHopCount() << ". hop" << endl;
	    }
	    if (!leafSet->mergeState(stReceived.back(),
			stateReceivedProximities.back()))
	    {
		EV << "Pastry: Error initializing own leafSet while "
		    "joining! Restarting ..." << endl;
		changeState(INIT);
		changeState(JOIN);
		return;
	    }
	    if (debugOutput)
	    {
		EV << "Pastry: [JOIN] initializing RoutingTable from all hops"
		    << endl;
	    }
	    if (!routingTable->initStateFromMsgVector(
			stReceived, stateReceivedProximities))
	    {
		EV << "Pastry: Error initializing own routingTable "
		    "while joining! Restarting ..." << endl;
		changeState(INIT);
		changeState(JOIN);
		return;
	    }
	    lastStateChange = now;
	    newLeafs();
	    changeState(READY);
	}
	else
	{
	    // process next state message in vector:
	    stateCache = *(stReceived.begin() +
		    stateReceivedProximities.size());
	    pingNodes();
	}
    }
    else
    {
	// state == READY

	if (debugOutput)
	{
	    EV << "Pastry: handling STATE message" << endl;
	    EV << "        type: " <<
		((stateCache->getPastryStateMsgType() == PASTRY_STATE_UPDATE)?
		 "update":"standard") << endl;
	    if (stateCache->getPastryStateMsgType() == PASTRY_STATE_UPDATE)
	    {
		EV << "        msg timestamp:      " <<
		    stateCache->getTimestamp() << endl;
		EV << "        last state change:  " <<
		    lastStateChange << endl;
	    }
	}
	
	if (stateCache->getPastryStateMsgType() == PASTRY_STATE_REPAIR)
	{
	    // try to repair own states based on repair message:
	    const TransportAddress& askRt =
		    routingTable->repair(stateCache, proxCache);
	    if (! askRt.isUnspecified()) sendRepairRequest(askRt);
	    
	    const TransportAddress& askLs =
		    leafSet->repair(stateCache, proxCache);
	    if (! askLs.isUnspecified()) sendRepairRequest(askLs);

	    // while not really known, it's safe to assume that a repair
	    // message changed our state:
	    lastStateChange = now;
	    newLeafs();
	}
	else if ((stateCache->getPastryStateMsgType() == PASTRY_STATE_UPDATE)
		&& (stateCache->getTimestamp() < lastStateChange))
	{
	    // if we received an update based on our outdated state,
	    // send new state message:
	    EV << "Pastry: outdated state from " << stateCache->getSender()
		<< endl;
	    sendStateTables(stateCache->getSender());
	}
	else
	{
	    // merge info in own state tables:
	    if (neighborhoodSet->mergeState(stateCache, proxCache))
		lastStateChange = now;

	    if (leafSet->mergeState(stateCache, proxCache))
		lastStateChange = now;
	    
	    if (routingTable->mergeState(stateCache, proxCache))
		lastStateChange = now;

	    newLeafs();
	}

	updateTooltip();
	
	// if state message was not an update, send one back:
	if (stateCache->getPastryStateMsgType() != PASTRY_STATE_UPDATE)
	{
	    sendStateTables(stateCache->getSender(), PASTRY_STATE_UPDATE,
		    stateCache->getTimestamp());
	}
	
	if (debugOutput)
	{
	    EV << "Pastry::checkProxCache() deleting STATE message "
		<< static_cast<void*>(stateCache) << endl;
	}
	delete stateCache;
	stateCache = NULL;
	proxCache.pr_rt.clear();

	// process next queued message:
	if (! stateCacheQueue.empty())
	{
	    stateCache = (PastryStateMessage*) stateCacheQueue.pop();
	    pingNodes();
	}
    }
}

void Pastry::handleRpcTimeout(BaseCallMessage* msg,
	const TransportAddress& dest, int rpcId)
{
    const NodeHandle* node;
    int i;
    int n;
    PastryPingCacheEntry pce;
    std::map<TransportAddress, BaseRouteMessage*>::iterator pos;
    
    // update ping cache:
    pce.inserted = simTime();
    pce.rtt = PASTRY_PROX_INFINITE;
    pingCache[dest.ip] = pce;

    RPC_SWITCH_START(msg)
	RPC_ON_CALL(Ping)
	{
	    EV << "Pastry: Ping timeout occured." << endl;

	    if (msg->isName("PING ring check"))
	    {
		// direct neighbor on the ring found dead:
		handleFailedNode(dest);
		break;
	    }

	    if (stateCache && msg->isName("PING received state"))
	    {
		// look for node in routing table of processed state message:
		n = stateCache->getRoutingTableArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getRoutingTable(i));
		    if ((!node->isUnspecified()) && (node->ip == dest.ip))
			*(proxCache.pr_rt.begin() + i) = PASTRY_PROX_INFINITE;
		}
		
		// look for node in leaf set of processed state message:
		n = stateCache->getLeafSetArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getLeafSet(i));
		    if ((!node->isUnspecified()) && (node->ip == dest.ip))
			*(proxCache.pr_ls.begin() + i) = PASTRY_PROX_INFINITE;
		}

		// look for node in neighborhood set of processed state message:
		n = stateCache->getNeighborhoodSetArraySize();
		for (i = 0; i < n; i++)
		{
		    node = &(stateCache->getNeighborhoodSet(i));
		    if ((!node->isUnspecified()) && (node->ip == dest.ip))
			*(proxCache.pr_ns.begin() + i) = PASTRY_PROX_INFINITE;
		}
		checkProxCache();
	    }

	    if (msg->isName("PING next hop"))
	    {
		// handle failed node:
		handleFailedNode(dest);

		// this could initiate a re-join, exit the handler in that
		// case because all local data was erased:
		if (state != READY) break;

		// handle forward queue entry
		if ((pos = recFwdQueue.find(dest)) == recFwdQueue.end()) break;
		BaseRouteMessage* rmsg = pos->second;
		recFwdQueue.erase(pos);
		sendToKey(rmsg->getDestKey(), rmsg);
	    }

	    break;
	}
    RPC_SWITCH_END()
}

void Pastry::sendRepairRequest(const TransportAddress& ask)
{
    if (ask.isUnspecified())
	opp_error("Pastry::sendRepairRequest(): asked for repair from "
		"unspecified node!");
    
    PastryRepairRequestMessage* msg = new PastryRepairRequestMessage("RREQ");
    msg->setPastryMsgType(PASTRY_MSG_RREQ);
    msg->setSignaling(true);
    msg->setSendStateTo(thisNode);
    msg->setLength(PASTRYRREQ_L(msg));
    RECORD_STATS(repairReqSent++; repairReqBytesSent += msg->byteLength());
    sendMessageToUDP(ask, msg);
}

bool Pastry::handleFailedNode(const TransportAddress& failed)
{
    if (state != READY)
    {
	return false;
    }

    if (failed.isUnspecified())
	opp_error("Pastry::handleFailedNode(): failed is unspecified!");

    const TransportAddress& lsAsk = leafSet->failedNode(failed);
    const TransportAddress& rtAsk = routingTable->failedNode(failed);
    neighborhoodSet->failedNode(failed);

    if (! lsAsk.isUnspecified()) sendRepairRequest(lsAsk);
    if (! rtAsk.isUnspecified()) sendRepairRequest(rtAsk);

    newLeafs();

    if (lsAsk.isUnspecified() && (! leafSet->isValid()))
    {
	EV << "Pastry: lost connection to the network, trying to re-join."
	    << endl;
	reJoin();
    }

    return true;
}

void Pastry::sendStateTables(const TransportAddress& destination, int type, ...)
{
    int hops = 0;
    bool last = false;
    simtime_t timestamp = 0;

    if ((type == PASTRY_STATE_JOIN) || (type == PASTRY_STATE_UPDATE))
    {
	// additional parameters needed:
	va_list ap;
	va_start(ap, type);
	if (type == PASTRY_STATE_JOIN)
	{
	    hops = va_arg(ap, int);
	    last = static_cast<bool>(va_arg(ap, int));
	}
	else
	{
	    timestamp = va_arg(ap, simtime_t);
	}
	va_end(ap);
    }
    
    // create new state message and fill in standard content:
    PastryStateMessage* stateMsg = new PastryStateMessage("STATE");
    stateMsg->setPastryMsgType(PASTRY_MSG_STATE);
    stateMsg->setSignaling(true);
    stateMsg->setPastryStateMsgType(type);
    stateMsg->setSender(thisNode);
    routingTable->dumpToStateMessage(stateMsg);
    leafSet->dumpToStateMessage(stateMsg);
    neighborhoodSet->dumpToStateMessage(stateMsg);

    // special fields for some types:
    if (type == PASTRY_STATE_JOIN)
    {
	stateMsg->setJoinHopCount(hops);
	stateMsg->setLastHop(last);
    }
    
    if (type == PASTRY_STATE_UPDATE)
    {
	EV << "Pastry: sending state (update) to " << destination << endl;
	stateMsg->setTimestamp(timestamp);
    }
    else
    {
	EV << "Pastry: sending state (standard) to " << destination << endl;
	stateMsg->setTimestamp(simTime());
    }
    
    stateMsg->setLength(PASTRYSTATE_L(stateMsg));
    RECORD_STATS(stateSent++; stateBytesSent += stateMsg->byteLength());
    sendMessageToUDP(destination, stateMsg);
}

bool Pastry::isResponsible(const OverlayKey& key)
{
    if (leafSet->isClosestNode(key)) return true;
    const NodeHandle& dest = leafSet->getDestinationNode(key);
    return (! dest.isUnspecified());
}

void Pastry::handleAppMessage(BaseOverlayMessage* msg)
{
    delete msg;
}

void Pastry::updateTooltip()
{
    if (ev.isGUI()) {
	std::stringstream ttString;

	// show our predecessor and successor in tooltip
	ttString << leafSet->getPredecessor() << endl << thisNode << endl
	    << leafSet->getSuccessor();

        parentModule()->parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());

	// draw arrows:
	showOverlayNeighborArrow(leafSet->getSuccessor(), true,
		"m=m,50,0,50,0;o=red,1");
	showOverlayNeighborArrow(leafSet->getPredecessor(), false,
		"m=m,50,100,50,100;o=green,1");

    }
}

void Pastry::receiveChangeNotification(int category, cPolymorphic* details)
{
    // get new ip address
    thisNode.ip = IPAddressResolver().addressOf(
	parentModule()->parentModule()).get4();
}

void Pastry::finishOverlay()
{
    // collect statistics

    recordScalar("Pastry: Successful Joins", joins);
    recordScalar("Pastry: overall join tries", joinTries);
    recordScalar("Pastry: JOIN Messages seen", joinSeen);
    recordScalar("Pastry: bytes of JOIN Messages seen", joinBytesSeen);
    recordScalar("Pastry: JOIN Messages received", joinReceived);
    recordScalar("Pastry: bytes of JOIN Messages received",
	    joinBytesReceived);
    recordScalar("Pastry: JOIN Messages sent", joinSent);
    recordScalar("Pastry: bytes of JOIN Messages sent", joinBytesSent);
    recordScalar("Pastry: STATE Messages sent", stateSent);
    recordScalar("Pastry: bytes of STATE Messages sent", stateBytesSent);
    recordScalar("Pastry: STATE Messages received", stateReceived);
    recordScalar("Pastry: bytes of STATE Messages received",
	    stateBytesReceived);
    recordScalar("Pastry: REPAIR Requests sent", repairReqSent);
    recordScalar("Pastry: bytes of REPAIR Requests sent",
	    repairReqBytesSent);
    recordScalar("Pastry: REPAIR Requests received", repairReqReceived);
    recordScalar("Pastry: bytes of REPAIR Requests received",
	    repairReqBytesReceived);
    recordScalar("Pastry: STATE Requests sent", stateReqSent);
    recordScalar("Pastry: bytes of STATE Requests sent", stateReqBytesSent);
    recordScalar("Pastry: STATE Requests received", stateReqReceived);
    recordScalar("Pastry: bytes of STATE Requests received",
	    stateReqBytesReceived);

    // destroy self timer messages
    if (joinTimeout->isScheduled()) cancelEvent(joinTimeout);
    if (readyWait->isScheduled()) cancelEvent(readyWait);
    if (secondStageWait->isScheduled()) cancelEvent(secondStageWait);
    if (ringCheck->isScheduled()) cancelEvent(ringCheck);

    delete joinTimeout;
    delete readyWait;
    delete secondStageWait;
    delete ringCheck;

    purgeVectors();

    // remove this node from the bootstrap list
    if (!thisNode.key.isUnspecified()) bootstrapOracle->removePeer(thisNode);
}

NodeVector* Pastry::findNode(const OverlayKey& key, BaseOverlayMessage* msg)
{
    NodeVector* nextHop = new NodeVector(1);
    const NodeHandle* next;
    PastryFindNodeExtData* findNodeExt = NULL;
    if (msg && msg->hasObject("findNodeExt"))
    {
	findNodeExt = check_and_cast<PastryFindNodeExtData*>(
		msg->getObject("findNodeExt"));
    }

    if (state != READY)
    {
	return nextHop;
    }
    else if (key.isUnspecified() || leafSet->isClosestNode(key))
    {
	nextHop->add(thisNode);
	if ((! key.isUnspecified()) && (findNodeExt))
	{
	    // final hop of a JOIN message, send state tables:
	    const TransportAddress& stateRecipient =
		findNodeExt->getSendStateTo();
	    if (!stateRecipient.isUnspecified())
	    {
		sendStateTables(stateRecipient, PASTRY_STATE_JOIN,
			findNodeExt->getJoinHopCount(), true);
	    }
	}
    }
    else
    {
	// Send state tables on any JOIN message we see:
	if (findNodeExt)
	{
	    const TransportAddress& stateRecipient =
		findNodeExt->getSendStateTo();
	    if (!stateRecipient.isUnspecified())
	    {
		RECORD_STATS(joinSeen++);
		sendStateTables(stateRecipient, PASTRY_STATE_JOIN,
			findNodeExt->getJoinHopCount(), false);
	    }
	}

	next = &(leafSet->getDestinationNode(key));

	if (next->isUnspecified())
	    next = &(routingTable->lookupNextHop(key));

	if (next->isUnspecified())
	{
	    // call findCloserNode() on all state objects

	    if (optimizeLookup)
	    {
		const NodeHandle* tmp;
		next = &(routingTable->findCloserNode(key, true));
		tmp = &(neighborhoodSet->findCloserNode(key, true));

		if ((! tmp->isUnspecified()) &&
			(leafSet->isCloser(*tmp, key, *next)))
		{
		    next = tmp;
		}

		tmp = &(leafSet->findCloserNode(key, true));
		if ((! tmp->isUnspecified()) &&
			(leafSet->isCloser(*tmp, key, *next)))
		{
		    next = tmp;
		}
	    }
	    else
	    {
		next = &(routingTable->findCloserNode(key));

		if (next->isUnspecified())
		    next = &(neighborhoodSet->findCloserNode(key));

		if (next->isUnspecified())
		    next = &(leafSet->findCloserNode(key));
	    }
	}

	if (!next->isUnspecified())
	{
	    if (findNodeExt)
	    {
		findNodeExt->setJoinHopCount(
			findNodeExt->getJoinHopCount() + 1);
	    }
	    nextHop->add(*next);
	}
    }
    
    return nextHop;
}

AbstractLookup* Pastry::createLookup(const BaseRouteMessage* msg)
{
    PastryFindNodeExtData* findNodeExt =
	new PastryFindNodeExtData("findNodeExt");

    if (msg)
    {
	const PastryMessage* pmsg = dynamic_cast<const PastryMessage*>(
		msg->encapsulatedMsg());
	if ((pmsg) && (pmsg->getPastryMsgType() == PASTRY_MSG_JOIN))
	{
	    const PastryJoinMessage* jmsg =
		check_and_cast<const PastryJoinMessage*>(pmsg);
	    findNodeExt->setSendStateTo(jmsg->getSendStateTo());
	    findNodeExt->setJoinHopCount(1);
	}
    }
    findNodeExt->setLength(PASTRYFINDNODEEXTDATA_L);
    AbstractLookup* newLookup = new BaseLookup(this,
	    baseLookupConfig, findNodeExt);
    lookups.insert(newLookup);
    delete findNodeExt;
    return newLookup;
}

bool stateMsgIsSmaller(const PastryStateMessage* msg1,
	const PastryStateMessage* msg2)
{
    return (msg1->getJoinHopCount() < msg2->getJoinHopCount());
}

std::ostream& operator<<(std::ostream& os, const PastryStateMsgProximity pr)
{
    os << "PastryStateMsgProximity {" << endl;
    os << "  pr_rt {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_rt.begin();
	    i != pr.pr_rt.end(); ++i)
    {
	os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "  pr_ls {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_ls.begin();
	    i != pr.pr_ls.end(); ++i)
    {
	os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "  pr_ns {" << endl;
    for (std::vector<simtime_t>::const_iterator i = pr.pr_ns.begin();
	    i != pr.pr_ns.end(); ++i)
    {
	os << "    " << *i << endl;
    }
    os << "  }" << endl;
    os << "}" << endl;
    return os;
}

