//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file PastryLeafSet.cc
 * @author Felix Palmen
 */

#include "PastryLeafSet.h"
#include "PastryTypes.h"

Define_Module(PastryLeafSet);

void PastryLeafSet::earlyInit(void)
{
    WATCH_VECTOR(leaves);
}

void PastryLeafSet::initializeSet(uint numberOfLeaves, double repairTimeout,
	const NodeHandle& owner)
{
    if (numberOfLeaves % 2) throw "numberOfLeaves must be even.";
    
    this->owner = owner;
    this->numberOfLeaves = numberOfLeaves;
    this->repairTimeout = repairTimeout;
    
    if (!leaves.empty()) leaves.clear();

    // fill Set with unspecified node handles
    for (uint i = numberOfLeaves; i>0; i--)
	leaves.push_back(NodeHandle::UNSPECIFIED_NODE);

    // initialize iterators to mark the beginning of bigger/smaller keys
    // in the set
    bigger = leaves.begin() + (numberOfLeaves >> 1);
    smaller = bigger - 1;

    // reset repair marker:
    if (!awaitingRepair.empty()) awaitingRepair.clear();

    newLeafs = false;
}

const NodeHandle& PastryLeafSet::getSuccessor(void) const
{
    return *bigger;
}

const NodeHandle& PastryLeafSet::getPredecessor(void) const
{
    return *smaller;
}

bool PastryLeafSet::isValid(void) const
{
    return (!(smaller->isUnspecified() || bigger->isUnspecified()));
}

const NodeHandle& PastryLeafSet::getDestinationNode(
	const OverlayKey& destination)
{
    std::vector<NodeHandle>::const_iterator i;
    const OverlayKey* smallest;
    const OverlayKey* biggest;
    const NodeHandle* ret = &NodeHandle::UNSPECIFIED_NODE;

    // check whether destination is inside leafSet:

    smallest = &(getSmallestKey());
    biggest = &(getBiggestKey());
    if (smallest->isUnspecified()) smallest = &(owner.key);
    if (biggest->isUnspecified()) biggest = &(owner.key);

    if (!destination.isBetweenLR(*smallest, *biggest)) return *ret;
    
    // find the closest node:

    for (i = leaves.begin(); i != leaves.end(); i++)
    {
	if (i->isUnspecified()) continue;

	// note for next line:
	// * dereferences iterator, & gets address of element.
	if (isCloser(*i, destination, *ret)) ret = &(*i);
    }

    return *ret;
}

bool PastryLeafSet::isClosestNode(const OverlayKey& destination) const
{
    // check for simple cases first
    if (owner.key == destination)
    {
	return true;
    }

    if (bigger->isUnspecified() && smaller->isUnspecified())
    {
	return true;
    }

    // check if the next bigger or smaller node in the set is closer
    // than own node
    bool biggerIsCloser = false;
    bool smallerIsCloser = false;

    if (! bigger->isUnspecified())
    {
	biggerIsCloser = isCloser(*bigger, destination);
    }
    if (! smaller->isUnspecified())
    {
	smallerIsCloser = isCloser(*smaller, destination);
    }
    
    // return true if both are not closer
    return ((!biggerIsCloser) && (!smallerIsCloser));
}

void PastryLeafSet::dumpToStateMessage(PastryStateMessage* msg) const
{
    uint i = 0;
    std::vector<NodeHandle>::const_iterator it;

    msg->setLeafSetArraySize(numberOfLeaves);
    for (it = leaves.begin(); it != leaves.end(); it++)
	msg->setLeafSet(i++, *it);
}

bool PastryLeafSet::mergeNode(const NodeHandle& node, simtime_t prox)
{
    std::vector<NodeHandle>::iterator it;
    bool ret = false;
    bool seenoverflow = false;
    const OverlayKey* last = &(owner.key);

    // look for correct position in "smaller" half of leafset
    it = smaller;
    while (true)
    {
	if (it->key.isUnspecified())
	{
	    // end of smaller entries found
	    if (!seenoverflow || (node.key < *last))
	    {
		// new entry can be put here if
		// - the key is smaller than the last seen in leafset
		// - or the 0 was not crossed yet
		insertLeaf(it, node);
		ret = true;
	    }
	    break;
	}
	else if (it->key == node.key)
	{
	    // found the same entry, exit loop without doing anything
	    break;
	}
	else if (node.key.isBetween(it->key, *last))
	{
	    // found correct position for inserting the new entry between
	    // existing ones
	    insertLeaf(it, node);
	    ret = true;
	    break;
	}
	
	// continue processing the loop
	else if (it->key > *last) seenoverflow = true;
	if (it == leaves.begin()) break;
	last = &(it->key);
	--it;
    }

    // look for correct position in "bigger" half of leafset,
    // details are the same as above.
    last = &(owner.key);
    seenoverflow = false;
    it = bigger;

    while (it != leaves.end())
    {
	if (it->key.isUnspecified())
	{
	    if (!seenoverflow || (node.key > *last))
	    {
		insertLeaf(it, node);
		ret = true;
	    }
	    break;
	}
	else if (it->key == node.key)
	{
	    break;
	}
	else if (node.key.isBetween(*last, it->key))
	{
	    insertLeaf(it, node);
	    ret = true;
	    break;
	}
	else if (it->key < *last) seenoverflow = true;
	last = &(it->key);
	++it;
    }

    return ret;
}

void PastryLeafSet::insertLeaf(std::vector<NodeHandle>::iterator& it,
	const NodeHandle& node)
{
    bool issmaller = (it <= smaller);
    if (issmaller)
    {
	leaves.insert(++it, node);
	leaves.erase(leaves.begin());
    }
    else
    {
	leaves.insert(it, node);
	leaves.pop_back();
    }
    newLeafs = true;
    bigger = leaves.begin() + (numberOfLeaves >> 1);
    smaller = bigger - 1;
}

void PastryLeafSet::dumpToVector(std::vector<TransportAddress>& affected) const
{
    std::vector<NodeHandle>::const_iterator it;

    for (it = leaves.begin(); it != leaves.end(); it++)
	if (!it->isUnspecified())
	    affected.push_back(*it);
}

const NodeHandle& PastryLeafSet::getSmallestNode(void) const
{
    std::vector<NodeHandle>::const_iterator i = leaves.begin();
    while ((i->isUnspecified()) && (i != smaller)) i++;
    return *i;
}

const OverlayKey& PastryLeafSet::getSmallestKey(void) const
{
    return getSmallestNode().key;
}

const NodeHandle& PastryLeafSet::getBiggestNode(void) const
{
    std::vector<NodeHandle>::const_iterator i = leaves.end()-1;
    while ((i->isUnspecified()) && (i != bigger)) i--;
    return *i;
}

const OverlayKey& PastryLeafSet::getBiggestKey(void) const
{
    return getBiggestNode().key;
}

const NodeHandle& PastryLeafSet::findCloserNode(const OverlayKey& destination,
	bool optimize)
{
    std::vector<NodeHandle>::const_iterator i;
    const NodeHandle* ret = &NodeHandle::UNSPECIFIED_NODE;

    // this will only be called after getDestinationNode() returned
    // NodeHandle::UNSPECIFIED_NODE, so a closer Node can only be the biggest
    // or the smallest node in the LeafSet.

    const NodeHandle& smallest = getSmallestNode();
    const NodeHandle& biggest = getBiggestNode();

    if ((!smallest.isUnspecified()) &&
	    (specialCloserCondition(smallest, destination, *ret)))
    {
	if (optimize) ret = &smallest;
	else return smallest;
    }

    if ((!biggest.isUnspecified()) &&
	    (specialCloserCondition(biggest, destination, *ret)))
    {
	if (optimize) ret = &biggest;
	else return biggest;
    }

    return *ret;
}

const TransportAddress& PastryLeafSet::failedNode(
	const TransportAddress& failed)
{
    std::vector<NodeHandle>::iterator i;
    const TransportAddress* ask;
    bool left = true;

    // search failed node in leafset:
    for (i = leaves.begin(); i != leaves.end(); i++)
    {
	if (i == bigger) left = false;
	if ((! i->isUnspecified()) && (i->ip == failed.ip)) break;
    }

    // failed node not in leafset:
    if (i == leaves.end()) return TransportAddress::UNSPECIFIED_NODE;

    // remove failed node:
    leaves.erase(i);
    newLeafs = true;

    // insert UNSPECIFIED_NODE at front or back and return correct node
    // to ask for repair:
    if (left)
    {
	leaves.insert(leaves.begin(), NodeHandle::UNSPECIFIED_NODE);
	bigger = leaves.begin() + (numberOfLeaves >> 1);
	smaller = bigger - 1;
	ask = static_cast<const TransportAddress*>(&(getSmallestNode()));
    }
    else
    {
	leaves.push_back(NodeHandle::UNSPECIFIED_NODE);
	bigger = leaves.begin() + (numberOfLeaves >> 1);
	smaller = bigger - 1;
	ask = static_cast<const TransportAddress*>(&(getBiggestNode()));
    }

    if (! ask->isUnspecified())
	awaitingRepair[*ask] = PLSRepairData(simTime(), left);
    
    return *ask;
}

const TransportAddress& PastryLeafSet::repair(const PastryStateMessage* msg,
	const PastryStateMsgProximity& prox)
{
    std::map<TransportAddress, PLSRepairData>::iterator it;
    const TransportAddress* ask;
    bool left;

    simtime_t now = simTime();

    // first eliminate outdated entries in awaitingRepair:
    for (it = awaitingRepair.begin(); it != awaitingRepair.end();)
    {
	if (it->second.ts < (now - repairTimeout))
	    awaitingRepair.erase(it++);
	else it++;
    }
    
    // don't expect any more repair messages:
    if (awaitingRepair.empty()) return TransportAddress::UNSPECIFIED_NODE;

    // look for source node in our list:
    if ( (it = awaitingRepair.find(msg->getSender())) == awaitingRepair.end() )
	return TransportAddress::UNSPECIFIED_NODE;

    // which side of the LeafSet is affected:
    left = it->second.left;
    
    // remove source node from list:
    awaitingRepair.erase(it);

    // merge info from repair message:
    if (mergeState(msg, prox))
    {
	EV << "Pastry: LeafSet repair was successful." << endl;
	return TransportAddress::UNSPECIFIED_NODE;
    }
    else
    {
	// repair did not succeed, try again:
	ask = &( left ? getSmallestNode() : getBiggestNode() );
	if (ask->isUnspecified())
	{
	    EV << "Pastry: LeafSet giving up repair attempt." << endl;
	}
	else
	{
	    awaitingRepair[*ask] = PLSRepairData(simTime(), left);
	}
	return *ask;
    }
}

PastryNewLeafsMessage* PastryLeafSet::getNewLeafsMessage(void)
{
    std::vector<NodeHandle>::const_iterator it;
    PastryNewLeafsMessage* msg;
    uint i = 0;
    
    if (! newLeafs) return NULL;
    newLeafs = false;

    msg = new PastryNewLeafsMessage("PastryNewLeafs");
    
    msg->setLeafsArraySize(numberOfLeaves);
    for (it = leaves.begin(); it != leaves.end(); it++)
	msg->setLeafs(i++, *it);

    msg->setLength(PASTRYNEWLEAFS_L(msg));
    return msg;
}
