//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file PastryRoutingTable.cc
 * @author Felix Palmen
 */

#include "PastryRoutingTable.h"

Define_Module(PastryRoutingTable);

inline uint32_t PastryRoutingTable::digitAt(uint n,
	const OverlayKey& key) const
{
    return key.get(OverlayKey::getLength() - ++n*bitsPerDigit, bitsPerDigit);
}

void PastryRoutingTable::earlyInit(void)
{
    WATCH_VECTOR(rows);
}

void PastryRoutingTable::initializeTable(uint bitsPerDigit,
	double repairTimeout, const NodeHandle& owner)
{
    this->owner = owner;
    this->repairTimeout = repairTimeout;
    this->bitsPerDigit = bitsPerDigit;
    nodesPerRow = 1 << bitsPerDigit; // 2 ^ bitsPerDigit

    // forget old routing table contents in case of restart:
    if (!rows.empty()) rows.clear();

    // clear pending repair requests:
    if (!awaitingRepair.empty()) awaitingRepair.clear();

    // Create first row in table:
    addRow();
}

const PastryExtendedNode& PastryRoutingTable::nodeAt(uint row, uint col) const
{
    if (rows.size() <= row) return unspecNode;
    if (col >= nodesPerRow) return unspecNode;

    return *((rows.begin()+row)->begin()+col);
}

const NodeHandle& PastryRoutingTable::lookupNextHop(
	const OverlayKey& destination)
{
    uint shl = owner.key.sharedPrefixLength(destination) / bitsPerDigit;
    uint digit = digitAt(shl, destination);

    if (shl >= rows.size())
    {
	EV << "Pastry: Unable to find next hop for " << destination
	    << ", row is empty." << endl;
	return NodeHandle::UNSPECIFIED_NODE;
    }

    const PastryExtendedNode& next = nodeAt(shl, digit);

    if (next.node.isUnspecified())
    {
	EV << "Pastry: Unable to find next hop for " << destination <<
	    ", routing table entry is empty." << endl;
    }
    return next.node;
}

const NodeHandle& PastryRoutingTable::findCloserNode(
    const OverlayKey& destination, bool optimize)
{
    const PastryExtendedNode* entry;

    if (optimize)
    {
	// pointer to later return value, initialize to unspecified, so
	// the specialCloserCondition() check will be done against our own
	// node as long as no node closer to the destination than our own was
	// found.
	const NodeHandle* ret = &NodeHandle::UNSPECIFIED_NODE;

	// a numerically closer node can only be found in the row containing
	// nodes with the same prefix length and in the row above.
	int shl = owner.key.sharedPrefixLength(destination) / bitsPerDigit;
	int digit = digitAt(shl, destination);
	int x = digitAt(shl, owner.key); // position index of own node

	// first try the row with same prefix length:
	int n = nodesPerRow;
	int a = digit - 1; // position index of search to the left
	int b = digit + 1; // position index of search to the right

	while ((a >= 0) || (b < n))
	{
	    // no need to continue search in one direction when own entry is
	    // reached:
	    if (a == x) a = -1;
	    if (b == x) b = n;

	    if (a >= 0)
	    {
		entry = &(nodeAt(shl, a));
		if ((!entry->node.isUnspecified()) &&
			specialCloserCondition(entry->node, destination, *ret))
		    ret = &(entry->node);
		a--;
	    }
	    if (b < n)
	    {
		entry = &(nodeAt(shl, b));
		if ((!entry->node.isUnspecified()) &&
			specialCloserCondition(entry->node, destination, *ret))
		    ret = &(entry->node);
		b++;
	    }
	}

	// it this was not the first row, two more nodes to check:
	if (shl != 0)
	{
	    // go up one row:
	    x = digitAt(--shl, owner.key);

	    if (destination < owner.key)
	    {
		entry = &(nodeAt(shl, digit - 1));
		if ((!entry->node.isUnspecified()) && 
			specialCloserCondition(entry->node, destination, *ret))
		    ret = &(entry->node);
	    }
	    else
	    {
		entry = &(nodeAt(shl, digit + 1));
		if ((!entry->node.isUnspecified()) && 
			specialCloserCondition(entry->node, destination, *ret))
		    ret = &(entry->node);
	    }
	}

	return *ret; // still unspecified if no closer node was found
    }
    else
    {
	// no optimization, return the first closer node found
	for (uint y = 0; y < rows.size(); ++y)
	{
	    for (uint x = 0; x < nodesPerRow; ++x)
	    {
		entry = &(nodeAt(y, x));
		if ((!entry->node.isUnspecified()) &&
			specialCloserCondition(entry->node, destination))
		    return entry->node;
	    }
	}

	return NodeHandle::UNSPECIFIED_NODE;
    }
}

void PastryRoutingTable::dumpToStateMessage(PastryStateMessage* msg) const
{
    uint i = 0;
    std::vector<PRTRow>::const_iterator itRows;
    PRTRow::const_iterator itCols;

    msg->setRoutingTableArraySize(rows.size() * nodesPerRow);

    for (itRows = rows.begin(); itRows != rows.end(); itRows++)
	for (itCols = itRows->begin(); itCols != itRows->end(); itCols++)
	    msg->setRoutingTable(i++, itCols->node);
}

bool PastryRoutingTable::mergeNode(const NodeHandle& node, simtime_t prox)
{
    uint shl;
    uint digit;
    PRTRow::iterator position;

    shl = owner.key.sharedPrefixLength(node.key) / bitsPerDigit;
    digit = digitAt(shl, node.key);

    while (rows.size() <= shl) addRow();
    position = (rows.begin() + shl)->begin() + digit;
    if (position->node.isUnspecified() || (prox < position->rtt))
    {
	EV << "Pastry: Node " << owner.key << endl;
	EV << "        placing node " << node.key << "in row "
	    << shl << ", col" << digit << endl;
	if (! position->node.isUnspecified())
	{
	    EV << "        (replaced because of better proximity: "
		<< prox << " < " << position->rtt << ")" << endl;
	}
	position->node = node;
	position->rtt = prox;
	return true;
    }
    return false;
}

bool PastryRoutingTable::initStateFromMsgVector(
	const std::vector<PastryStateMessage*>& messages,
	const std::vector<PastryStateMsgProximity>& proximities)
{
    std::vector<PastryStateMessage*>::const_iterator itMsg;
    std::vector<PastryStateMsgProximity>::const_iterator itProx;
    PastryStateMessage* msg;
    int hopCheck = 0;

    itProx = proximities.begin();
    for (itMsg = messages.begin(); itMsg != messages.end(); itMsg++)
    {
	msg = *itMsg;
	if (msg->getJoinHopCount() != ++hopCheck) return false;
	mergeState(msg, *(itProx++));
    }
    return true;
}

void PastryRoutingTable::dumpToVector(std::vector<TransportAddress>& affected)
    const
{
    std::vector<PRTRow>::const_iterator itRows;
    PRTRow::const_iterator itCols;

    for (itRows = rows.begin(); itRows != rows.end(); itRows++)
	for (itCols = itRows->begin(); itCols != itRows->end(); itCols++)
	    if (!itCols->node.isUnspecified())
		affected.push_back(itCols->node);
}

void PastryRoutingTable::addRow(void)
{
    PRTRow row(nodesPerRow, unspecNode);

    // place own node at correct position:
    (row.begin() + digitAt(rows.size(), owner.key))->node = owner;
    rows.push_back(row);
}

std::ostream& operator<<(std::ostream& os, const PRTRow& row)
{
    os << "Pastry RoutingTable row {" << endl;
    for (PRTRow::const_iterator i = row.begin(); i != row.end(); i++)
	os << "        " << i->node << " ; Ping: " << i->rtt << endl;
    os << "    }";
    return os;
}

const TransportAddress& PastryRoutingTable::failedNode(
	const TransportAddress& failed)
{
    std::vector<PRTRow>::iterator itRows;
    PRTRow::iterator itCols;
    PRTTrackRepair tmpTrack;
    
    bool found = false;

    // find node in table:
    for (itRows = rows.begin(); itRows != rows.end(); itRows++)
    {
	for (itCols = itRows->begin(); itCols != itRows->end(); itCols++)
	{
	    if ((! itCols->node.isUnspecified()) &&
		    (itCols->node.ip == failed.ip))
	    {
		itCols->node = NodeHandle::UNSPECIFIED_NODE;
		itCols->rtt = PASTRY_PROX_UNDEF;
		found = true;
		break;
	    }
	}
	if (found) break;
    }

    // not found, nothing to do:
    if (!found) return TransportAddress::UNSPECIFIED_NODE;

    // else fill temporary record:
    tmpTrack.failedRow = itRows - rows.begin();
    tmpTrack.failedCol = itCols - itRows->begin();
    tmpTrack.node = TransportAddress::UNSPECIFIED_NODE;
    findNextNodeToAsk(tmpTrack);
    tmpTrack.timestamp = simTime();

    if (tmpTrack.node.isUnspecified())
	return TransportAddress::UNSPECIFIED_NODE;
    awaitingRepair.push_back(tmpTrack);
    return awaitingRepair.back().node;
}

const TransportAddress& PastryRoutingTable::repair(
	const PastryStateMessage* msg, const PastryStateMsgProximity& prox)
{
    std::vector<PRTTrackRepair>::iterator it;
    simtime_t now = simTime();

    // first eliminate outdated entries in awaitingRepair:
    for (it = awaitingRepair.begin(); it != awaitingRepair.end();)
    {
	if (it->timestamp < (now - repairTimeout))
	    it = awaitingRepair.erase(it);
	else it++;
    }

    // don't expect any more repair messages:
    if (awaitingRepair.empty()) return TransportAddress::UNSPECIFIED_NODE;

    // look for source node in our list:
    for (it = awaitingRepair.begin(); it != awaitingRepair.end(); it++)
	if (it->node == msg->getSender()) break;

    // if not found, return from function:
    if (it == awaitingRepair.end()) return TransportAddress::UNSPECIFIED_NODE;

    // merge state:
    mergeState(msg, prox);

    // repair not yet done?
    if (nodeAt(it->failedRow, it->failedCol).node.isUnspecified())
    {
	// ask next node
	findNextNodeToAsk(*it);
	if (it->node.isUnspecified())
	{
	    // no more nodes to ask, give up:
	    EV << "Pastry: RoutingTable giving up repair attempt." << endl;
	    awaitingRepair.erase(it);
	    return TransportAddress::UNSPECIFIED_NODE;
	}
	else return it->node;
    }
    
    // repair done: clean up
    EV << "Pastry: RoutingTable repair was successful." << endl;
    awaitingRepair.erase(it);
    return TransportAddress::UNSPECIFIED_NODE;
}

void PastryRoutingTable::findNextNodeToAsk(PRTTrackRepair& track) const
{
    const TransportAddress* ask;

    if (track.node.isUnspecified())
    {
	track.askedRow = track.failedRow;
	if (track.failedCol == 0)
	    track.askedCol = 1;
	else
	    track.askedCol = 0;
	ask = static_cast<const TransportAddress*>(
		&(nodeAt(track.askedRow, track.askedCol).node));
	track.node = *ask;
	if (! track.node.isUnspecified()) return;
    }

    do
    {
	// point to next possible position in routing table:
	track.askedCol++;
	if ((track.askedRow == track.failedRow) &&
		(track.askedCol == track.failedCol)) track.askedCol++;
	if (track.askedCol == nodesPerRow)
	{
	    if ((track.askedRow > track.askedCol) ||
		    (track.askedRow == (rows.size() - 1)))
	    {
		// no more nodes that could be asked, give up:
		track.node = TransportAddress::UNSPECIFIED_NODE;
		return;
	    }
	    track.askedRow++;
	    track.askedCol = 0;
	}

	ask = static_cast<const TransportAddress*>(
		&(nodeAt(track.askedRow, track.askedCol).node));
	track.node = *ask;
    }
    while (track.node.isUnspecified());
}

