#ifndef __PASTRYROUTINGTABLE_H_
#define __PASTRYROUTINGTABLE_H_

//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file PastryRoutingTable.h
 * @author Felix Palmen
 */

#include <vector>

#include <omnetpp.h>

#include <NodeHandle.h>

#include "PastryStateObject.h"
#include "PastryTypes.h"
#include "PastryMessage_m.h"

/**
 * Vector-type of a line in Pastry RoutingTable
 */
typedef std::vector<PastryExtendedNode> PRTRow;

/**
 * Struct for tracking attempts to repair a routing table entry
 */
struct PRTTrackRepair
{
    TransportAddress node;//*< the node last asked for repair
    simtime_t timestamp;  //*< the time when it was asked
    uint failedRow;       //*< row of the failed node
    uint failedCol;	  //*< col of the failed node
    uint askedRow;	  //*< row of the node last asked
    uint askedCol;	  //*< col of the node last asked
};

/**
 * Routing table module
 *
 * This module contains the routing table of the Chord implementation.
 *
 * @author Felix Palmen
 * @see Pastry
 */
class PastryRoutingTable : public PastryStateObject
{

  public:
    /**
     * Initializes the routing table. This should be called on startup
     *
     * @param bitsPerDigit Pastry configuration parameter
     * @param repairTimeout Pastry configuration parameter
     * @param owner the node this table belongs to
     */
    void initializeTable(uint bitsPerDigit, double repairTimeout,
	    const NodeHandle& owner);

    /**
     * gets the next hop according to the Pastry routing scheme.
     *
     * @param destination the destination key
     * @return the NodeHandle of the next Node or NodeHandle::UNSPECIFIED_NODE
     *	  if no next hop could be determined
     */
    const NodeHandle& lookupNextHop(const OverlayKey& destination);

    /**
     * try to find a node numerically closer to a given key with the same
     * shared prefix as the current node in the routing table. this method
     * is to be called, when a regular next hop couldn't be found or wasn't
     * reachable.
     *
     * @param destination the destination key
     * @param optimize if set, check all nodes and return the best/closest one
     * @return a closer NodeHandle or NodeHandle::UNSPECIFIED_NODE if none was
     *	  found
     */
    virtual const NodeHandle& findCloserNode(const OverlayKey& destination,
	    bool optimize = false);

    /**
     * tells the routing table that a node has failed
     *
     * @param failed the failed node
     * @return a node to ask for REPAIR or TransportAddress::UNSPECIFIED_NODE
     */
    virtual const TransportAddress& failedNode(const TransportAddress& failed);

    /**
     * attempt to repair the routing using a received REPAIR message
     *
     * @param msg the state message of type REPAIR
     * @param prox record of proximity values matching the state message
     * @return another node to ask for REPAIR or
     *    TransportAddress::UNSPECIFIED_NODE
     */
    virtual const TransportAddress& repair(const PastryStateMessage* msg,
	    const PastryStateMsgProximity& prox);

    /**
     * dump content of the table to a PastryStateMessage
     *
     * @param msg the PastryStateMessage to be filled with entries
     */
    virtual void dumpToStateMessage(PastryStateMessage* msg) const;

    /**
     * merge a node in the RoutingTable
     *  
     * @param node the node to merge
     * @param prox proximity value of the node
     * @return true if node was merged
     */
    bool mergeNode(const NodeHandle& node, simtime_t prox);
  
    /**
     * initialize table from vector of PastryStateMessages received during
     * JOIN phase. The vector has to be sorted by JoinHopCount of the messages
     *
     * @param messages the vector of PastryStateMessages
     * @param proximities a vector of corresponding proximity tables
     * @return true on success
     */
    bool initStateFromMsgVector(
	    const std::vector<PastryStateMessage*>& messages,
	    const std::vector<PastryStateMsgProximity>& proximities);

    /**
     * appends all routing table entries to a given vector of
     * TransportAddresses, needed to find all Nodes to be notified after
     * joining.
     *
     * @param affected the vector to fill with routing table entries
     */
    virtual void dumpToVector(std::vector<TransportAddress>& affected) const;

  private:
    uint bitsPerDigit;
    double repairTimeout;
    uint nodesPerRow;
    std::vector<PRTRow> rows;
    std::vector<PRTTrackRepair> awaitingRepair;

    virtual void earlyInit(void);

    /**
     * adds a new line to the routing table
     */
    void addRow(void);

    /**
     * returns n'th pastry digit from a key
     *
     * @param n which digit to return
     * @param key extract digit from this key
     * @return a pastry digit
     */
    uint32_t digitAt(uint n, const OverlayKey& key) const;

    /**
     * returns routing table entry at specified position
     * 
     * @param row the number of the row
     * @param col the number of the column
     */
    const PastryExtendedNode& nodeAt(uint row, uint col) const;

    /**
     * helper function, updates a PRTTrackRepair structure to point to the next
     * node that can be asked for repair
     *
     * @param track the PRTTrackRepair structure
     */
    void findNextNodeToAsk(PRTTrackRepair& track) const;

};

/**
 * Stream output operator to make WATCH() do something useful with the
 * routing table
 */
std::ostream& operator<<(std::ostream& os, const PRTRow& row);

#endif
