#!/usr/bin/perl -w
use strict;
use POSIX;
use Getopt::Std;
use Term::Cap;
use Cwd 'abs_path';

=head1 NAME

runSims.pl - Script for running multiple OverSim simulations with different
parameters

=head1 SYNOPSIS

./runSims.pl [-r{s}[:{e}]]

without arguments, runsims.ini is parsed, the total amount of
different configurations is calculated and those values are displayed.

To actually simulate something, the parameter -r must be given with a
single configuration number between 1 and the total number of
different configurations, or with a range of configurations in the
format {s}:{e} where {s} is the first and {e} the last configuration
to be simulated.

=head1 EXAMPLE

./runSims.pl -r5:8

    will run simulations for the configurations number 5 - 8.

=head1 FILES

./runsims.ini

    Template for the generated omnetpp.ini files and runSims.pl
    configuration. See example file for syntax.

./sim{x}/

    For each configuration to be simulated, a sim{x} directory with the
    number of the configuration {x} will be created. runSims.pl will not
    overwrite existing directories, so make sure they do not exist.

=head1 FUNCTIONS

=over 4

=cut

# global variables
# --

my $term;		# for storing the needed termcaps
my $config = {
  cmdline => {},	# command line arguments
  path => "",		# path to OverSim binary
  simtime => 0,		# number of simsecs for each simulation
  repeatruns => 1,	# run each simulation this many times
  startrun => 0,	# first config number to be simulated
  endrun => 0,		# last config number to be simulated
  numruns => 0,		# total number of configurations
  head => "",		# string for omnetpp.ini lines above the runs
  foot => "",		# string for omnetpp.ini lines below the runs
  default => {},	# the default values for each run
  variations => [],	# the configuration variations to be combined
};
my $runs = [];		# array of all resulting configurations


# functions
# --

=item printusage()

show the user a short "Usage" message and exit unsuccessful

=cut

sub printusage
{
  print STDERR "Usage:\n";
  print STDERR "    $ARGV[0]          - to check configuration\n";
  print STDERR "    $ARGV[0] -rs[:e]  - to actually simulate the distinct ".
      "runs s [to e]\n";
  exit (1);
}


=item fetchconfig()

this function reads configuration from the commandline and from runsims.ini
and fills the global $config hash accordingly.

=cut

sub fetchconfig
{
  # Read command line options
  getopts("r:", $config->{'cmdline'}) or usage();

  # Configurations to simulate given?
  if ($config->{'cmdline'}->{r} &&
      ($config->{'cmdline'}->{r} =~ m/^\s*(\d+)\s*(:\s*(\d+)\s*)?$/))
  {
    $config->{'startrun'} = $1;
    if ($3)
    {
      $config->{'endrun'} = $3;
    }
    else
    {
      $config->{'endrun'} = $1;
    }
  }

  # Read from runsims.ini
  open CONFIG, "<runsims.ini" or die "runsims.ini not found!";

  my $line;			# holds the line to be parsed
  my $variation;		# holds the current [Variation] section
  my $collect = "head";		# What is read right now (head, default,
  				#   variation, runsims, foot)

  while ($line = <CONFIG>)
  {

    # Check for new .ini-section
    if ($line =~ m/^\[(.+)\]/)
    {
      if ($1 eq "Defaults")
      {
        # Section with default values for all runs found
        $collect = "default";
	next;
      }
      elsif ($1 eq "Variation")
      {
        # A new Variation section
        $collect = "variations";
	$variation = {
	  count => 0,
	  options => []
	};
	push @{$config->{'variations'}}, $variation;
	next;
      }
      elsif ($1 eq "RunSims")
      {
        # Section with runSims.pl configuration found
        $collect = "runsims";
	next;
      }
    }
    if (($collect ne "head") && ($line =~ /^$/))
    {
      # Put anything not recognized after a blank line below the runs
      # in the resulting omnetpp.ini files
      $collect = "foot";
      next;
    }

    # Read runSims.pl configuration
    if ($collect eq "runsims")
    {
      if ($line =~ m/^\s*(.+?)\s*=\s*(.+?)\s*$/)
      {
        my $key = $1;
	my $val = $2;
	if ($key eq "path")
	{
	  # Path to OverSim binary
	  if ($val =~ m/^\s*"?(.+)"?\s*$/)
	  {
	    # make it an absolute path and check if it is really executable
	    $config->{'path'} = abs_path($1);
	    if (-x $config->{'path'})
	    {
	      print "[config] using OverSim: $config->{'path'}\n";
	    }
	    else
	    {
	      die ("``$val'' not found or not executable!");
	    }
	  }
	  else
	  {
	    die ("Invalid OverSim path: ``$val''!");
	  }
	}
	elsif ($key eq "simtime")
	{
	  # Simtime for each run, numerical values in h, m and s are
	  # recognized
	  if ($val =~ m/^(\d+(\.\d+)?)\s*h/)
	  {
	    my $hours = $1;
	    print "[config] Simulations will run for $hours hours (";
	    $config->{'simtime'} = int(($hours * 3600) + .5);
	    print "$config->{'simtime'} seconds)\n";
	  }
	  elsif ($val =~ m/^(\d+(\.\d+)?)\s*m/)
	  {
	    my $minutes = $1;
	    print "[config] Simulations will run for $minutes minutes (";
	    $config->{'simtime'} = int(($minutes * 60) + .5);
	    print "$config->{'simtime'} seconds)\n";
	  }
	  elsif ($val =~ m/^(\d+)\s*s?/)
	  {
	    # If there is no unit given, seconds are assumed. No fractions
	    # are allowed for seconds.
	    my $seconds = int($1);
	    print "[config] Simulations will run for $seconds seconds\n";
	    $config->{'simtime'} = $seconds;
	  }
	  else
	  {
	    die ("Illegal format of simulation time option: ``$val''");
	  }
	}
	elsif ($key eq "repeatruns")
	{
	  # How often to repeat the same run
	  if ($val !~ m/^\s*\d+\s*$/)
	  {
	    die ("Illegal repeatruns value in RunSims section: ``$val''");
	  }
	  $config->{'repeatruns'} = int($val);
	  if (! $config->{'repeatruns'})
	  {
	    die ("Illegal repeatruns value in RunSims section: ``$val''");
	  }
	  print "[config] The same run will be repeated ".
	      "$config->{'repeatruns'} times\n";
	}
      }
      else
      {
        die ("malformed configuration line in RunSims section: ".
	    "``$line''");
      }
    }
    elsif ($collect eq "head")
    {
      # lines to be put above the runs
      if ($line =~ m/^include (.*)$/)
      {
        # make "include" paths absolute
        $line = "include ".abs_path($1);
      }
      $config->{'head'} .= $line;
    }
    elsif ($collect eq "foot")
    {
      # lines to be put below the runs
      if ($line =~ m/^include (.*)$/)
      {
        # make "include" paths absolute
        $line = "include ".abs_path($1);
      }
      $config->{'foot'} .= $line;
    }
    elsif ($line =~ m/^\s*(.+?)\s*=\s*(.+?)\s*$/)
    {
      # Found a configuration value in .INI format
      my $key = $1;
      my $val = $2;
      if ($collect eq "default")
      {
        # add to default run configuration
        $config->{'default'}->{$key} = $val;
      }
      else
      {
        # if this is inside a Variation section, allow multiple values
	# separated by ','. All options inside one Variation must have the
	# same number of values.
      	my @options = split(/,/, $val);
	if (! $variation->{'count'})
	{
	  $variation->{'count'} = @options;
	  for (@options)
	  {
	    push @{$variation->{'options'}}, { $key => $_ };
	  }
	}
	elsif ($variation->{'count'} != scalar(@options))
	{
	  die ("Configuration error near ``$line'': Inside one Variation, ".
	      "all options must have the same number of values!");
	}
	else
	{
	  for (my $i=0; $i<$variation->{'count'}; ++$i)
	  {
	    $variation->{'options'}->[$i]->{$key} = $options[$i];
	  }
	}
      }
    }
  }
  close(CONFIG);

  # Sanity checks - some values must be present
  if (!$config->{'path'})
  {
    die ("Configuration Error: no 'path' given in RunSims section!");
  }
  if (!$config->{'simtime'})
  {
    die ("Configuration Error: no 'simtime' given in RunSims section!");
  }

  # set sim-time-limit as default for all runs:
  $config->{'default'}->{'sim-time-limit'} = $config->{'simtime'}."s";

  # determine the total number of different configurations
  $config->{'numruns'} = 1;
  for (@{$config->{'variations'}})
  {
    $config->{'numruns'} *= $_->{count};
  }
  print "[config] Have $config->{'numruns'} different configurations.\n";

  # check sanity of start and end configuration given on commandline
  if ($config->{'startrun'})
  {
    if (($config->{'startrun'} > $config->{'numruns'}) ||
        ($config->{'endrun'} > $config->{'numruns'}) ||
	($config->{'startrun'} > $config->{'endrun'}))
    {
      die ("Illegal range of runs given: ".
          "``$config->{'startrun'}:$config->{'endrun'}''!");
    }
    print "[config] Attempt to simulate runs $config->{'startrun'} ".
        "till $config->{'endrun'}.\n";
  }
}

=item getTerm()

determine the termcaps needed for run()

=cut

sub getTerm
{
  # get needed termcap entries:
  my $termios = new POSIX::Termios;
  $termios->getattr;
  my $terminal = Term::Cap->Tgetent( { TERM => undef,
      OSPEED => $termios->getospeed } );
  my $term = {
    "clear" => $terminal->Tputs('cl'),
    "c_show" => $terminal->Tputs('ve'),
    "c_hide" => $terminal->Tputs('vi'),
  };
  return $term;
}

=item cleanup()

do what is needed to reset the terminal

=cut

sub cleanup
{
  print $term->{'c_show'};
}

=item run()

execute one configuration [numberofruns] times, move omnetpp.vec out of the
way each time.

while running, progress information is shown constantly.

Parameters:

B<cfgnum>: The current configuration number (for progress indicators)

B<cfgtotal>: Total number of configurations (for progress indicators)

=cut

sub run
{
  my $i;		# loop variable
  my $cfgnum = shift;	# current config number
  my $cfgtotal = shift; # total number of configs

  # pre-calculate some values for progress indicators
  my $cfgfrac = (1.0 / $cfgtotal);
  my $runfrac = (1.0 / $config->{'repeatruns'});
  my $cfgprog = (($cfgnum - 1) / $cfgtotal);
  my $runprog;

  # repeat the run as requested
  for ($i=1; $i<=$config->{'repeatruns'}; ++$i)
  {
    $runprog = (($i - 1) / $config->{'repeatruns'});
    print $term->{'c_hide'};
    open OVERSIM, "$config->{'path'} -r$i -uCmdenv 2>oversim-error.log |";
    my $line;
    my $status;
    my $elapsed;
    my $prog;
    while ($line = <OVERSIM>)
    {
      # Read status from OverSim process
      if ($line =~ m/.*Event.*Elapsed: ([0-9\.]+)s/)
      {
        $status = $line;
        $status .= <OVERSIM>;
        $status .= <OVERSIM>;
        $elapsed = $1;
	$prog = $elapsed / $config->{simtime};
        print $term->{'clear'};
        print "\n    OverSim running\n";
	print "     configuration:    $cfgnum / $cfgtotal\n";
	print "     run:              $i / $config->{'repeatruns'}\n\n";
        print " --------\n";
        print "   status\n";
        print $status;
        print " --------\n\n";
        printf("Progress | this run: %.2f %% | this conf: %.2f %% | ".
	    "overall: %.2f %%\n",
            100 * $prog, 100 * ($runprog + $runfrac * $prog),
	    100 * ($cfgprog + $cfgfrac * $runprog
	    		+ $cfgfrac * $runfrac * $prog));
      }
    }
    close OVERSIM;
    if ($?)
    {
      # An error occured:
      cleanup();
      print "\n\n** Error in configuration $cfgnum, run $i:\n";
      open ERROR, "<oversim-error.log";
      while (<ERROR>)
      {
        print;
      }
      close ERROR;
      chdir("..");
      exit ($?);
    }

    # move omnetpp.vec out of the way
    rename("omnetpp.vec", "omnetpp.$i.vec");
    (-s "oversim-error.log") or unlink("oversim-error.log");
  }
  cleanup();
}

=item makeruns()

calculate all the different configurations, works recursively

=cut

sub makeruns
{
  my $variations = shift;
  my $run = shift;
  if (! ref($variations))
  {
    # first call, copy variations array from global config:
    my @variations = @{$config->{'variations'}};
    $variations = \@variations;
  }
  if (! ref($run))
  {
    # first call, run is just the defaults section:
    $run = $config->{'default'};
  }

  # shift first variation out of the array
  my $variation = shift @$variations;

  # loop over all options in the variation
  my $opts;
  for $opts (@{$variation->{'options'}})
  {
    # copy the run
    my %run = %$run;

    # and change it according to the variation option
    for (keys(%$opts))
    {
      if ($opts->{$_})
      {
      	$run{$_} = $opts->{$_};
      }
      else
      {
        delete $run{$_};
      }
    }

    if (@$variations)
    {
      # continue with next variation
      makeruns($variations, \%run);
    }
    else
    {
      # a leaf of this recursive tree is one permutation, save in $runs array
      push @$runs, \%run;
    }
  }

  # put current variation back in the array before going up the tree
  unshift @$variations, $variation;
}

=item runall()

loop over all configurations, create the sim{x} directory and omnetpp.ini
file and execute the simulation with run()

=cut

sub runall
{
  my $i;
  my $j;

  my $n = $config->{'endrun'} - $config->{'startrun'} + 1;
  my $x = 1;
  for ($i = $config->{'startrun'}; $i <= $config->{'endrun'}; ++$i)
  {
    mkdir("sim$i") or die ("can't create directory sim$i");
    chdir("sim$i");
    open(INI, ">omnetpp.ini");
    print INI $config->{'head'};
    for ($j = 1; $j <= $config->{'repeatruns'}; ++$j)
    {
      print INI "[Run $j]\n";
      for (keys(%{$runs->[$i]}))
      {
        print INI "$_ = $runs->[$i]->{$_}\n";
      }
      print INI "\n";
    }
    print INI $config->{'foot'};
    close(INI);
    run($x,$n);
    chdir("..");
    ++$x;
  }
}

=item simabort()

signal handler for SIGINT and SIGTERM to cleanup the terminal in any case

=cut

sub simabort
{
  print "\n\n** Aborted!\n";
  cleanup();
  exit(1);
}
$SIG{'INT'} = \&simabort;
$SIG{'TERM'} = \&simabort;

=back

=cut

# main script
# --

$term = getTerm();
print "\n";
fetchconfig();
if ($config->{'startrun'})
{
  makeruns();
  runall();
}
else
{
  print "\n\n to actually start some of these runs, use -rs[:e]\n";
  print " for example -r1:4 to simulate the first 4 runs.\n\n";
}

=head1 AUTHOR

written by Felix M. Palmen <fmp@palmen.homeip.net>

